import sys

import matplotlib.pyplot as plt
import seaborn as sns
import warnings
import numpy as np

warnings.filterwarnings("ignore", category=RuntimeWarning)


def plot_loss(total_loss_train, total_loss_test):
    img = plt.figure()
    plt.plot(total_loss_train)
    plt.plot(total_loss_test)
    plt.ylim(0)
    plt.xticks((np.arange(0, len(total_loss_train), step=1)))
    plt.title('Model Loss')
    plt.ylabel('Loss')
    plt.xlabel('Epochs')
    plt.legend(['train', 'validation'], loc='upper left')
    return img



import torch
from torch.utils.data import Dataset


def train(epoch, tokenizer, model, device, loader, val_loader, optimizer, testing=True):
    model.train()
    running_loss = 0.0
    for _, data in enumerate(loader, 0):
        y = data['target_ids'].to(device, dtype=torch.long)
        y_ids = y[:, :-1].contiguous()
        lm_labels = y[:, 1:].clone().detach()
        lm_labels[y[:, 1:] == tokenizer.pad_token_id] = -100
        ids = data['source_ids'].to(device, dtype=torch.long)
        mask = data['source_mask'].to(device, dtype=torch.long)

        outputs = model(input_ids=ids,
                        attention_mask=mask,
                        decoder_input_ids=y_ids,
                        lm_labels=lm_labels)
        loss = outputs[0]

        running_loss += loss.item()

        optimizer.zero_grad()
        loss.backward()
        optimizer.step()

    running_loss_val = 0.0
    if testing:
        with torch.no_grad():
            for _, data in enumerate(val_loader, 0):
                y = data['target_ids'].to(device, dtype=torch.long)
                y_ids = y[:, :-1].contiguous()
                lm_labels = y[:, 1:].clone().detach()
                lm_labels[y[:, 1:] == tokenizer.pad_token_id] = -100
                ids = data['source_ids'].to(device, dtype=torch.long)
                mask = data['source_mask'].to(device, dtype=torch.long)

                outputs = model(
                    input_ids=ids,
                    attention_mask=mask,
                    decoder_input_ids=y_ids,
                    lm_labels=lm_labels)
                loss = outputs[0]

                running_loss_val += loss.item()

        return running_loss / len(loader), running_loss_val / len(val_loader)
    else:
        return running_loss / len(loader), 0


class CustomDataset(Dataset):

    def __init__(self, dataframe, tokenizer, source_len, summ_len):
        self.tokenizer = tokenizer
        self.data = dataframe
        self.source_len = source_len
        self.summ_len = summ_len
        self.text = self.data.text
        self.ctext = self.data.ctext

    def __len__(self):
        return len(self.text)

    def __getitem__(self, index):
        ctext = str(self.ctext[index])
        ctext = ' '.join(ctext.split())

        text = str(self.text[index])
        text = ' '.join(text.split())

        source = self.tokenizer.batch_encode_plus([ctext],
                                                  max_length=self.source_len,
                                                  pad_to_max_length=True,
                                                  return_tensors='pt')
        target = self.tokenizer.batch_encode_plus([text],
                                                  max_length=self.summ_len,
                                                  pad_to_max_length=True,
                                                  return_tensors='pt')

        source_ids = source['input_ids'].squeeze()
        source_mask = source['attention_mask'].squeeze()
        target_ids = target['input_ids'].squeeze()
        target_mask = target['attention_mask'].squeeze()

        return {
            'source_ids': source_ids.to(dtype=torch.long),
            'source_mask': source_mask.to(dtype=torch.long),
            'target_ids': target_ids.to(dtype=torch.long),
            'target_ids_y': target_ids.to(dtype=torch.long)
        }


def inference(tokenizer, model, device, loader):
    model.eval()
    predictions = []
    actuals = []
    with torch.no_grad():
        for _, data in enumerate(loader, 0):
            y = data['target_ids'].to(device, dtype=torch.long)
            ids = data['source_ids'].to(device, dtype=torch.long)
            mask = data['source_mask'].to(device, dtype=torch.long)

            generated_ids = model.generate(
                input_ids=ids,
                attention_mask=mask,
                max_length=150,
                num_beams=2,
                repetition_penalty=2.5,
                length_penalty=1.0,
                early_stopping=True
            )
            preds = [
                tokenizer.decode(
                    g,
                    skip_special_tokens=True,
                    clean_up_tokenization_spaces=True) for g in generated_ids]
            target = [
                tokenizer.decode(
                    t,
                    skip_special_tokens=True,
                    clean_up_tokenization_spaces=True)for t in y]

            predictions.extend(preds)
            actuals.extend(target)
    return predictions, actuals



import os
import shutil
import cv2
import glob
import pandas as pd
from libra.data_generation.dataset_labelmatcher import get_similar_column
from libra.data_generation.grammartree import get_value_instruction

# Preprocesses images from images to median of heighs/widths


def setwise_preprocessing(data_path, new_folder, height, width):
    training_path = data_path + "/training_set"
    testing_path = data_path + "/testing_set"

    heights = []
    widths = []

    # first dictionary is training and second is testing set
    paths = [training_path, testing_path]
    dict = []
    data_size = []
    num_classes = [0, 0]
    for index in range(2):
        info = process_class_folders(paths[index])
        data_size.append(len(info[0]))
        heights += info[0]
        widths += info[1]
        if info[2] < 2:
            raise BaseException(
                f"Directory: {paths[index]} contains {info[2]} classes. Need at least two classification folders.")
        num_classes[index] += info[2]
        for key, value in info[3].items():
            if len(value) == 0:
                raise BaseException(
                    f"Class: {key} contans {len(value)} images. Need at least one image in this class.")
        dict.append(info[3])

    if num_classes[0] != num_classes[1]:
        raise BaseException(
            f"Number of classes in testing_set and training_set are not equal. Training set has {num_classes[0]} and testing set has {num_classes[1]}")

    classification = num_classes[0]
    height1, width1 = calculate_medians(heights, widths)
    if height is None:
        height = height1
    if width is None:
        width = width1

    # resize images
    for index, p in enumerate(paths):
        for class_folder, images in dict[index].items():
            for image_name, image in images.items():
                dict[index][class_folder][image_name] = process_color_channel(
                    image, height, width)
        if new_folder:
            folder_names = ["proc_training_set", "proc_testing_set"]
            create_folder(data_path, folder_names[index])
            for class_folder, images in dict[index].items():
                add_resized_images(
                    data_path + "/" + folder_names[index],
                    class_folder,
                    images)
        else:
            for class_folder, images in dict[index].items():
                replace_images(p + "/" + class_folder, images)

    return {"num_categories": classification,
            "height": height,
            "width": width,
            "train_size": data_size[0],
            "test_size": data_size[1]}


# processes a csv_file containing image paths and creates a testing/training
# folders with resized images inside the same directory as the csv file
def csv_preprocessing(csv_file,
                      data_path,
                      instruction,
                      image_column,
                      training_ratio,
                      height,
                      width):

    df = pd.read_csv(csv_file)
    if instruction is None:
        raise BaseException(
            "Instruction was not given for csv file to be processed.")

    label = get_similar_column(get_value_instruction(instruction), df)
    avoid_directories = ["proc_training_set", "proc_testing_set"]
    data_paths = [
        data_path +
        "/" +
        d for d in os.listdir(data_path) if os.path.isdir(
            data_path +
            "/" +
            d) and d not in avoid_directories]

    file_extensions = ["jpg", "jpeg", "png", "gif"]
    need_file_extension = False
    path_included = False

    count = 0
    while image_column is None:
        if count > 20:
            raise BaseException(
                f"Could not locate column containing image information.")
        count += 1
        random_row = df.sample()
        for column, value in random_row.iloc[0].items():
            if isinstance(value, str):
                if os.path.exists(data_path + "/" +
                                  (value if value[0] != "/" else value[1:])):
                    path_included = True
                    image_column = column
                    break
                # add file extension if not included
                if value.split(".")[-1] in file_extensions:
                    file = [value]
                else:
                    file = []
                    for extension in file_extensions:
                        file.append(value + "." + extension)

                # look through all data_paths for file
                for path in data_paths:
                    for file_option in file:
                        if os.path.exists(path + "/" + file_option):
                            if file_option.split(".")[-1] in file_extensions:
                                need_file_extension = True
                            image_column = column
                            break
            if image_column is not None:
                break

    else:
        if os.path.exists(data_path + "/" + df.iloc[0][image_column]):
            path_included = True
        elif df.iloc[0][image_column].split(".")[-1]:
            need_file_extension = True

    df = df[[image_column, label]].dropna()

    heights = []
    widths = []
    classifications = df[label].value_counts()
    if len(classifications) < 2:
        raise BaseException(
            f"{csv_file} contains {len(classifications)} classes. Need at least two classification labels.")
    for key, value in classifications.items():
        if value < 2:
            raise BaseException(
                f"Class: {key} contans {value} images. Need at least two images in this class.")

    image_list = []

    # get the median heights and widths
    for index, row in df.iterrows():
        if path_included:
            p = data_path + "/" + \
                (row[image_column][1:] if row[image_column][0] == "/" else row[image_column])
            img = cv2.imread(p)
        else:
            for path in data_paths:
                if need_file_extension:
                    for extension in file_extensions:
                        p = path + "/" + row[image_column] + "." + extension
                        img = cv2.imread(p)
                        if img is not None:
                            break
                else:
                    p = path + "/" + row[image_column]
                    img = cv2.imread(p)
                if img is not None:
                    break
        if img is None:
            raise BaseException(
                f"{row[image_column]} could not be found in any directories.")
        image_list.append(img)
        heights.append(img.shape[0])
        widths.append(img.shape[1])

    height1, width1 = calculate_medians(heights, widths)
    if height is None:
        height = height1
    if width is None:
        width = width1

    # create training and testing folders
    create_folder(data_path, "proc_training_set")
    create_folder(data_path, "proc_testing_set")
    # create classification folders
    for classification in classifications.keys():
        create_folder(data_path + "/proc_training_set", classification)
        create_folder(data_path + "/proc_testing_set", classification)

    data_size = [0, 0]
    class_count = dict.fromkeys(classifications.keys(), 0)

    # save images into correct folder
    for index, row in df.iterrows():
        # resize images
        img = process_color_channel(image_list[index], height, width)
        p = "proc_" + (os.path.basename(row[image_column])
                       if path_included else row[image_column])
        if need_file_extension:
            p += ".jpg"
        if class_count[row[label]] / \
                classifications[row[label]] < training_ratio:
            data_size[0] += 1
            class_count[row[label]] += 1
            save_image(data_path + "/proc_training_set", img, p, row[label])
        else:
            data_size[1] += 1
            class_count[row[label]] += 1
            save_image(data_path + "/proc_testing_set", img, p, row[label])

    return {"num_categories": len(classifications),
            "height": height,
            "width": width,
            "train_size": data_size[0],
            "test_size": data_size[1]}


# preprocesses images when given a folder containing class folders
def classwise_preprocessing(data_path, training_ratio, height, width):
    info = process_class_folders(data_path)
    if info[2] < 2:
        raise BaseException(
            f"Directory: {data_path} contains {info[2]} classes. Need at least two classification folders.")
    img_dict = info[3]
    for key, value in img_dict.items():
        if len(value) < 2:
            raise BaseException(
                f"Class: {key} contans {len(value)} images. Need at least two images in this class.")
    height1, width1 = calculate_medians(info[0], info[1])
    if height is None:
        height = height1
    if width is None:
        width = width1
    num_classifications = info[2]

    # create training and testing folders
    create_folder(data_path, "proc_training_set")
    create_folder(data_path, "proc_testing_set")

    # create classification folders
    for classification in img_dict.keys():
        create_folder(data_path + "/proc_training_set", classification)
        create_folder(data_path + "/proc_testing_set", classification)

    data_size = [0, 0]
    for class_folder, images in img_dict.items():
        count = 0
        for image_name, image in images.items():
            resized_img = process_color_channel(image, height, width)
            if count / len(images) < training_ratio:
                data_size[0] += 1
                save_image(data_path + "/proc_training_set",
                           resized_img,
                           "proc_" + image_name,
                           class_folder)
            else:
                data_size[1] += 1
                save_image(data_path + "/proc_testing_set",
                           resized_img,
                           "proc_" + image_name,
                           class_folder)
            count += 1

    return {"num_categories": num_classifications,
            "height": height,
            "width": width,
            "train_size": data_size[0],
            "test_size": data_size[1]}


# process a class folder by getting return a list of all the heights and widths
def process_class_folders(data_path):
    heights = []
    widths = []
    num_classifications = 0
    img_dict = {}
    avoid_directories = ["proc_training_set", "proc_testing_set"]

    for class_folder in os.listdir(data_path):
        if not os.path.isdir(
            data_path +
            "/" +
                class_folder) or class_folder in avoid_directories:
            continue

        folder_dict = {}
        folder_height = []
        folder_width = []

        for image in os.listdir(data_path + "/" + class_folder):
            if image == ".DS_Store":
                continue
            try:
                img = cv2.imread(data_path + "/" + class_folder + "/" + image)
                folder_height.append(img.shape[0])
                folder_width.append(img.shape[1])
                folder_dict[image] = img
            except BaseException:
                break
        else:
            heights += folder_height
            widths += folder_width
            img_dict[class_folder] = folder_dict
            num_classifications += 1

    return [heights, widths, num_classifications, img_dict]


# creates a folder with the given folder name and fills the folders
# with the images given
def add_resized_images(data_path, folder_name, images):
    # create processed folder
    os.mkdir(data_path + "/proc_" + folder_name)
    # add images to processed folder
    for img_name, img in images.items():
        cv2.imwrite(
            data_path +
            "/proc_" +
            folder_name +
            "/proc_" +
            img_name,
            img)


# writes an image into the given path
def replace_images(data_path, loaded_shaped):
    for img_name, img in loaded_shaped.items():
        cv2.imwrite(data_path + "/" + img_name, img)


# creates a folder in the given path with the given folder name
def create_folder(path, folder_name):
    folder_name = "/" + folder_name
    if os.path.isdir(path + folder_name):
        shutil.rmtree(path + folder_name)
    os.mkdir(path + folder_name)


# saves an image to the given path inside the classification folder
# specified with the img_name
def save_image(path, img, img_name, classification):
    cv2.imwrite(
        path + "/" + classification + "/" + img_name,
        img)

# calculates the medians of the given lists of height and widths


def calculate_medians(heights, widths):
    heights.sort()
    widths.sort()
    height = heights[int(len(heights) / 2)]
    width = widths[int(len(widths) / 2)]
    return height, width


# resizes the image with the given height and width
def process_color_channel(img, height, width):
    chanels = [chanel for chanel in cv2.split(img)]

    for index, chanel in enumerate(chanels):
        if chanel.shape[0] > height:
            chanel = cv2.resize(
                chanel,
                dsize=(
                    chanel.shape[1],
                    height),
                interpolation=cv2.INTER_CUBIC)
        else:
            chanel = cv2.resize(
                chanel,
                dsize=(
                    chanel.shape[1],
                    height),
                interpolation=cv2.INTER_AREA)
        if chanel.shape[1] > width:
            chanel = cv2.resize(
                chanel,
                dsize=(
                    width,
                    height),
                interpolation=cv2.INTER_CUBIC)
        else:
            chanel = cv2.resize(
                chanel,
                dsize=(
                    width,
                    height),
                interpolation=cv2.INTER_AREA)
        chanels[index] = chanel

    return cv2.merge(chanels)

# distinguishes the preprocess mode to do


def set_distinguisher(data_path, read_mode):
    if read_mode is not None:
        if read_mode == "setwise":
            if os.path.isdir(data_path +
                             "/training_set") and os.path.isdir(data_path +
                                                                "/testing_set"):
                return {"read_mode": "setwise"}
            raise BaseException(
                f"training_set or testing_set folder not in f{data_path}")
        elif read_mode == "classwise":
            return {"read_mode": "classwise"}
        elif read_mode == "csvwise":
            csv_path = glob.glob(data_path + "/*.csv")
            if len(csv_path) == 1:
                return {
                    "read_mode": "csvwise",
                    "csv_path": csv_path[0]}
            elif len(csv_path) > 1:
                raise BaseException(
                    f"Too many csv files in {data_path}: {[os.path.basename(path) for path in csv_path]}")
            else:
                raise BaseException(f"No csv file in {data_path}")
        else:
            raise BaseException(f"{read_mode}, is an invalid read mode.")

    # check if setwise
    if os.path.isdir(data_path +
                     "/training_set") and os.path.isdir(data_path +
                                                        "/testing_set"):
        return {"read_mode": "setwise"}

    # check if contains a csv file
    csv_path = glob.glob(data_path + "/*.csv")
    if len(csv_path) == 1:
        return {"read_mode": "csvwise", "csv_path": csv_path[0]}
    elif len(csv_path) > 1:
        raise BaseException(
            f"Too many csv files in directory: {[os.path.basename(path) for path in csv_path]}")

    return {"read_mode": "classwise"}


# get the dataset info for the CNN model
def already_processed(data_path):
    training_path = data_path + "/training_set"
    testing_path = data_path + "/testing_set"

    if not os.path.isdir(training_path) or not os.path.isdir(testing_path):
        raise BaseException(
            f"Missing training_set or testing_set folder in directory: {data_path}")

    paths = [training_path, testing_path]
    num_categories = [0, 0]
    sizes = [0, 0]
    for index, path in enumerate(paths):
        for directory in os.listdir(path):
            class_path = path + "/" + directory
            if os.path.isdir(class_path):
                num_categories[index] += 1
                images = [img for img in os.listdir(
                    class_path) if img != ".DS_Store"]
                if len(images) == 0:
                    raise BaseException(
                        f"Class: {directory} in {path} contains no images.")
                height, width, _ = cv2.imread(
                    class_path + "/" + images[0]).shape
                sizes[index] += len(images)
        if num_categories[index] < 2:
            raise BaseException(
                f"Directory: {path} contains {num_categories[index]} classes. Need at least two classification folders.")

    if num_categories[0] != num_categories[1]:
        raise BaseException(
            f"Number of classes in testing_set and training_set are not equal. Training set has {num_categories[0]} and testing set has {num_categories[1]}")

    return {"num_categories": num_categories[0],
            "height": height,
            "width": width,
            "train_size": sizes[0],
            "test_size": sizes[1]}



import os
import pandas as pd
import tensorflow as tf
from sklearn.model_selection import train_test_split
from tensorflow.python.client import device_lib
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'


class DataReader():
    def __init__(
            self,
            filepath,
            trim=False,
            trim_format='random',
            trim_ratio=0.20):
        '''
        Constructor for the DataReader class.

        The DataReader class creates a Pandas DataFrame object
        based off of the datset's file extension, the format of trimming that needs to be applied to the
        dataset (i.e. random sampling), and how much of the dataset needs to be trimmed.

        :param filepath: The file path to the dataset (str)
        :param trim: Whether the dataset should be trimmed or not (bool)
        :param trim_format: The format/type of trimming (str)
        :param trim_ratio: The proportion of the dataset that needs to be trimmed (float)
        '''
        self.filepath = filepath
        self.trim = trim
        self.trim_format = trim_format
        self.trim_ratio = trim_ratio

    def retrieve_file_size(self):
        '''
        Returns the size of the dataset in megabytes

        :return: the size of the dataset in megabytes (float)
        '''
        file_stats = os.stat(self.filepath)
        return (file_stats.st_size / (1024 ** 2))

    def retrieve_extension(self):
        '''
        Returns the dataset's file extension (i.e. .csv, .json, .xlsx, etc.)

        :return: the dataset's file extension (str)
        '''
        ext_index = self.filepath.rindex('.')
        return self.filepath[ext_index:]

    def data_generator(self):
        '''
        Creates a Pandas DataFrame object based off of the dataset's file extension, whether a GPU is available or not,
        whether the user wants the dataset to be trimmed or not, and the format of trimming specified by the user.

        NOTE: This function currently only supports .csv (comma-separated values file),
        .xlsx (Microsoft Excel Open XML Spreadsheet), and .json (JavaScript Object Notation) files

        If the user's device contains a GPU, the dataset won't be trimmed unless the user specifies so. If the user's
        device doesn't contain a GPU, the dataset will automatically be trimmed regardless of whether the user specified
        for the dataset to be trimmed or not in order to ensure efficient processing by the CPU.

        If the user doesn't specify a specific form of trimming they want to apply to the dataset, random sampling will
        be applied by default.

        If the user doesn't specify a proportion/ratio of how much of the dataset needs to be trimmed, 20% of the
        dataset will be trimmed by default.

        :return: The dataset after being trimmed/pre-processed (Pandas DataFrame)
        '''
        if self.retrieve_extension() == '.csv':
            df = pd.read_csv(self.filepath)
        elif self.retrieve_extension() == '.xlsx':
            df = pd.read_excel(self.filepath)
            for data in df:
                if df[data].dtype.name == 'int64':
                    df[data] = df[data].astype(float)
        elif self.retrieve_extension() == '.json':
            df = pd.read_json(self.filepath)

        if self.is_gpu_available() == False:
            self.trim = True
        if self.trim:
            if self.trim_format == 'random':
                df = df.sample(frac=(1.0 - self.trim_ratio))
            # elif self.trim_format == 'stratify':
            #     # ADD STRATIFYING TECHNIQUE HERE
            #     y = df[self.strat_col]
            #     del df[self.strat_col]
            #     print(df.shape)
            #     x_train, x_test, y_train, y_test = train_test_split(df, y, stratify=y, test_size=0.1, random_state=0)
            #
            #     df = pd.concat([x_test, y_test])
            #     print(df.shape)

        return df

    def get_available_gpus(self):
        '''
        Returns a list of available GPUs on the current device running the program

        :return: List of available GPUs on the current device (list of Strings)
        '''
        local_device_protos = device_lib.list_local_devices()
        return [device.name for device in local_device_protos
                if device.device_type == 'GPU']

    def is_gpu_available(self):
        '''
        Returns a boolean value representing whether the current device has a GPU or not

        :return: Whether the current device has a GPU or not (bool)
        '''
        return tf.test.gpu_device_name() != ''


import tensorflow as tf
import keras
from keras import Model
from keras.layers import LSTM, Embedding, TimeDistributed, Concatenate
from tensorflow.python.keras.layers import Dense, Input
import numpy as np
from keras.models import model_from_json
import os

# Creates a regression neural network


def get_keras_model_reg(dataset, i):
    size_max_network = 10
    # base model
    model = tf.keras.Sequential()
    model.add(
        Dense(
            dataset['train'].shape[1],
            input_dim=dataset['train'].shape[1],
            kernel_initializer='normal',
            activation='relu'))
    model.add(Dense(64, activation="relu"))

    # Adds values depending on what the i value is
    for x in range(i):
        if (i < 5):
            model.add(
                Dense(
                    i * 64,
                    kernel_initializer="normal",
                    activation="relu"))
        else:
            model.add(
                Dense(
                    i * 64,
                    kernel_initializer="normal",
                    activation="relu"))

    model.add(Dense(1, kernel_initializer="normal"))
    model.compile(loss='mean_squared_error', optimizer='adam')
    return model


# Creates a classification neural network


def get_keras_model_class(dataset, i, num_classes):
    size_max_network = 10
    # base model
    model = tf.keras.Sequential()
    model.add(
        Dense(
            dataset['train'].shape[1],
            input_dim=dataset['train'].shape[1],
            kernel_initializer='normal',
            activation='relu'))
    model.add(Dense(64, activation="relu"))

    # Adds values depending on what the i value is
    for x in range(i):
        if (i < 5):
            model.add(
                Dense(
                    i * 64,
                    kernel_initializer="normal",
                    activation="relu"))
        else:
            model.add(
                Dense(
                    i * 64,
                    kernel_initializer="normal",
                    activation="relu"))

    model.add(Dense(num_classes, activation="softmax"))
    model.compile(
        loss='categorical_crossentropy',
        optimizer='adam',
        metrics=['accuracy'])

    return model


# Sentiment analysis model
def get_keras_text_class(max_features, num_classes, lr):
    model = tf.keras.models.Sequential()
    model.add(Embedding(max_features, 128))
    model.add(LSTM(128, dropout=0.2, recurrent_dropout=0.2))
    model.add(Dense(num_classes, activation='softmax'))
    opt = keras.optimizers.Adam(learning_rate=lr)
    model.compile(loss='binary_crossentropy',
                  optimizer=opt,
                  metrics=['accuracy'])

    return model


def getKerasConvolutional():
    print("entered")


# Loading the user defined model given that the user provides the model path.
# The user must provide the model Directory.
def load_model(self, load_path):
    load_file = load_path[load_path.rindex("/") + 1:load_path.rindex(".")]
    # loading model json file
    json_file = open(load_path, 'r')
    json_model = json_file.read()
    json_file.close()
    loaded_model = model_from_json(json_model)
    # loading weights in the model
    loaded_model.load_weights(load_file + ".h5")
    print("Model loaded...")
    return loaded_model



import os
import numpy as np
from colorama import Fore, Style
import pandas as pd
import torch
from keras_preprocessing import sequence
from sklearn.model_selection import train_test_split
from tensorflow.python.keras.callbacks import EarlyStopping
from tensorflow.python.keras.saving.saved_model.json_utils import Encoder
from torch.utils.data import DataLoader
from transformers import T5Tokenizer, T5ForConditionalGeneration
import tensorflow as tf

from libra.data_generation.dataset_labelmatcher import get_similar_column
from libra.data_generation.grammartree import get_value_instruction
from libra.modeling.prediction_model_creation import get_keras_text_class
import libra.plotting.nonkeras_generate_plots
from libra.plotting.generate_plots import generate_classification_plots
from libra.preprocessing.NLP_preprocessing import get_target_values, text_clean_up, lemmatize_text, encode_text
from libra.preprocessing.huggingface_model_finetune_helper import CustomDataset, train, inference
from libra.preprocessing.image_caption_helpers import load_image, map_func, CNN_Encoder, RNN_Decoder, get_path_column, \
    generate_caption_helper
from libra.preprocessing.data_reader import DataReader

# Sentiment analysis predict wrapper
from libra.query.supplementaries import save
from libra.plotting.generate_plots import plot_loss

counter = 0

currLog = 0


def clearLog():
    global currLog
    global counter

    currLog = ""
    counter = 0


def logger(instruction, found=""):
    '''
    logging function that creates hierarchial display of the processes of
    different functions. Copied into different python files to maintain
    global variables.

    :param instruction: what you want to be displayed
    :param found: if you want to display something found like target column
    '''
    
    global counter
    if counter == 0:
        print((" " * 2 * counter) + str(instruction) + str(found))
    elif instruction == "->":
        counter = counter - 1
        print(Fore.BLUE + (" " * 2 * counter) +
              str(instruction) + str(found) + (Style.RESET_ALL))
    else:
        print((" " * 2 * counter) + "|- " + str(instruction) + str(found))
        if instruction == "done...":
            print("\n" + "\n")

    counter += 1


def classify_text(self, text):
    '''
    function to perform sentiment analysis text classification

    :param text: text sent in/from written query to be analyzed
    '''
    
    sentimentInfo = self.models.get("Text Classification")
    vocab = sentimentInfo["vocabulary"]
    # Clean up text
    text = lemmatize_text(text_clean_up([text]))
    # Encode text
    text = encode_text(vocab, text)
    text = sequence.pad_sequences(text, sentimentInfo["max_text_length"])
    model = sentimentInfo["model"]
    prediction = tf.keras.backend.argmax(model.predict(text))
    return sentimentInfo["classes"][tf.keras.backend.get_value(prediction)[0]]


# Sentiment analysis query
def text_classification_query(self, instruction, drop=None,
                              preprocess=True,
                              test_size=0.2,
                              random_state=49,
                              learning_rate=1e-2,
                              epochs=20,
                              monitor="val_loss",
                              batch_size=32,
                              max_text_length=200,
                              max_features=20000,
                              generate_plots=True,
                              save_model=False,
                              save_path=os.getcwd()):

    '''
    function to apply text classification algorithm for sentiment analysis
    :param many params: used to hyperparametrize the function.
    :return a dictionary object with all of the information for the algorithm.
    '''

    if test_size < 0:
        raise Exception("Test size must be a float between 0 and 1")

    if test_size >= 1:
        raise Exception(
            "Test size must be a float between 0 and 1 (a test size greater than or equal to 1 results in no training "
            "data)")

    if epochs < 1:
        raise Exception("Epoch number is less than 1 (model will not be trained)")

    if batch_size < 1:
        raise Exception("Batch size must be equal to or greater than 1")

    if max_text_length < 1:
        raise Exception("Max text length must be equal to or greater than 1")

    if save_model:
        if not os.path.exists(save_path):
            raise Exception("Save path does not exists")

    data = DataReader(self.dataset)
    data = data.data_generator()

    if preprocess:
        data.fillna(0, inplace=True)

    if drop is not None:
        data.drop(drop, axis=1, inplace=True)

    X, Y, target = get_target_values(data, instruction, "label")
    Y = np.array(Y)
    classes = np.unique(Y)

    logger("->", "Target Column Found: {}".format(target))

    vocab = {}
    if preprocess:
        logger("Preprocessing data")
        X = lemmatize_text(text_clean_up(X.array))
        vocab = X
        X = encode_text(X, X)

    X = np.array(X)

    model = get_keras_text_class(max_features, len(classes), learning_rate)
    logger("Building Keras LSTM model dynamically")

    X_train, X_test, y_train, y_test = train_test_split(
        X, Y, test_size=test_size, random_state=random_state)

    X_train = sequence.pad_sequences(X_train, maxlen=max_text_length)
    X_test = sequence.pad_sequences(X_test, maxlen=max_text_length)

    y_vals = np.unique(np.append(y_train, y_test))
    label_mappings = {}
    for i in range(len(y_vals)):
        label_mappings[y_vals[i]] = i
    map_func = np.vectorize(lambda x: label_mappings[x])
    y_train = map_func(y_train)
    y_test = map_func(y_test)

    logger("Training initial model")

    # early stopping callback
    es = EarlyStopping(
        monitor=monitor,
        mode='auto',
        verbose=0,
        patience=5)

    history = model.fit(X_train, y_train, validation_data=(X_test, y_test),
                        batch_size=batch_size,
                        epochs=epochs, callbacks=[es], verbose=0)

    logger("->", "Final training loss: {}".format(history.history["loss"][len(history.history["loss"]) - 1]))
    logger("->", "Final validation loss: {}".format(history.history["val_loss"][len(history.history["val_loss"]) - 1]))
    logger("->", "Final validation accuracy: {}".format(
        history.history["val_accuracy"][len(history.history["val_accuracy"]) - 1]))

    plots = {}
    if generate_plots:
        # generates appropriate classification plots by feeding all
        # information
        logger("Generating plots")
        plots = generate_classification_plots(
            history, X, Y, model, X_test, y_test)

    if save_model:
        save(model, save_model, save_path=save_path)

    logger("Storing information in client object under key 'Text Classification'")
    # storing values the model dictionary

    self.models["Text Classification"] = {"model": model,
                                          "classes": classes,
                                          "plots": plots,
                                          "target": Y,
                                          "vocabulary": vocab,
                                          "interpreter": label_mappings,
                                          "max_text_length": max_text_length,
                                          'test_data': {'X': X_test, 'y': y_test},
                                          'losses': {
                                              'training_loss': history.history['loss'],
                                              'val_loss': history.history['val_loss']},
                                          'accuracy': {
                                              'training_accuracy': history.history['accuracy'],
                                              'validation_accuracy': history.history['val_accuracy']}}
    clearLog()
    return self.models["Text Classification"]


# Document summarization predict wrapper
def get_summary(self, text):
    modelInfo = self.models.get("Document Summarization")
    model = modelInfo['model']
    model.eval()
    tokenizer = T5Tokenizer.from_pretrained("t5-small")
    df = pd.DataFrame({'text': [""], 'ctext': [text]})
    params = {
        'batch_size': 1,
        'shuffle': True,
        'num_workers': 0
    }
    loader = DataLoader(
        CustomDataset(
            df,
            tokenizer,
            modelInfo["max_text_length"],
            modelInfo["max_sum_length"]),
        **params)
    predictions, truth = inference(tokenizer, model, "cpu", loader)
    return predictions[0]


# Text summarization query
def summarization_query(self, instruction, preprocess=True,
                        drop=None,
                        epochs=10,
                        batch_size=32,
                        learning_rate=1e-4,
                        max_text_length=512,
                        max_summary_length=150,
                        test_size=0.2,
                        random_state=49,
                        gpu=False,
                        generate_plots=True,
                        save_model=False,
                        save_path=os.getcwd()):

    '''
    function to apply algorithm for text summarization 
    :param many params: used to hyperparametrize the function.
    :return a dictionary object with all of the information for the algorithm.
    '''
    

    if test_size < 0:
        raise Exception("Test size must be a float between 0 and 1")

    if test_size >= 1:
        raise Exception(
            "Test size must be a float between 0 and 1 (a test size greater than or equal to 1 results in no training "
            "data)")

    if max_text_length < 2 | max_summary_length < 2:
        raise Exception("Text and summary must be at least of length 2")

    if epochs < 1:
        raise Exception("Epoch number is less than 1 (model will not be trained)")

    if batch_size < 1:
        raise Exception("Batch size must be equal to or greater than 1")

    if max_text_length < 1:
        raise Exception("Max text length must be equal to or greater than 1")

    if max_summary_length < 1:
        raise Exception("Max summary length must be equal to or greater than 1")

    if save_model:
        if not os.path.exists(save_path):
            raise Exception("Save path does not exists")

    if test_size == 0:
        testing = False
    else:
        testing = True

    if gpu:
        device = "cuda"
    else:
        device = "cpu"

    data = DataReader(self.dataset)
    data = data.data_generator()

    if drop is not None:
        data.drop(drop, axis=1, inplace=True)

    if preprocess:
        data.fillna(0, inplace=True)

    logger("Preprocessing data...")

    X, Y, target = get_target_values(data, instruction, "summary")
    df = pd.DataFrame({'text': Y, 'ctext': X})
    logger("->", "Target Column Found: {}".format(target))

    torch.manual_seed(random_state)
    np.random.seed(random_state)

    tokenizer = T5Tokenizer.from_pretrained("t5-small")

    train_size = 1 - test_size
    train_dataset = df.sample(
        frac=train_size,
        random_state=random_state).reset_index(
        drop=True)

    logger("Establishing dataset walkers")
    training_set = CustomDataset(
        train_dataset, tokenizer, max_text_length, max_summary_length)

    if testing:
        val_dataset = df.drop(train_dataset.index).reset_index(drop=True)

        val_set = CustomDataset(
            val_dataset,
            tokenizer,
            max_text_length,
            max_summary_length)

        val_params = {
            'batch_size': batch_size,
            'shuffle': False,
            'num_workers': 0
        }
        val_loader = DataLoader(val_set, **val_params)
    else:
        val_loader = None

    train_params = {
        'batch_size': batch_size,
        'shuffle': True,
        'num_workers': 0
    }

    training_loader = DataLoader(training_set, **train_params)
    # used small model
    model = T5ForConditionalGeneration.from_pretrained("t5-small")
    model = model.to(device)

    optimizer = torch.optim.Adam(
        params=model.parameters(), lr=learning_rate)

    logger('Fine-Tuning the model on your dataset...')
    total_loss_train = []
    total_loss_val = []
    for epoch in range(epochs):
        loss_train, loss_val = train(
            epoch, tokenizer, model, device, training_loader, val_loader, optimizer, testing=testing)
        total_loss_train.append(loss_train)
        total_loss_val.append(loss_val)

    logger("->", "Final training loss: {}".format(loss_train))
    if testing:
        logger("->", "Final validation loss: {}".format(loss_val))
    else:
        logger("->", "Final validation loss: {}".format("0, No validation done"))

    plots = {}
    if generate_plots:
        logger("Generating plots")
        plots.update({"loss": libra.plotting.nonkeras_generate_plots.plot_loss(total_loss_train, total_loss_val)})

    if save_model:
        logger("Saving model")
        path = save_path + "DocSummarization.pth"
        torch.save(model, path)
        logger("->", "Saved model to disk as DocSummarization.pth")

    logger("Storing information in client object under key 'Document Summarization'")

    self.models["Document Summarization"] = {
        "model": model,
        "max_text_length": max_text_length,
        "max_sum_length": max_summary_length,
        "plots": plots,
        'losses': {'training_loss': loss_train,
                   'val_loss': loss_val}
    }
    clearLog()
    return self.models["Document Summarization"]


# Image Caption Generation Prediction
def generate_caption(self, image):
    '''
    wrapper function of image_caption_query to predict caption for image
    :param image: image to be analyzed
    '''
    modelInfo = self.models.get("Image Caption")
    decoder = modelInfo['decoder']
    encoder = modelInfo['encoder']
    tokenizer = modelInfo['tokenizer']
    image_features_extract_model = modelInfo['feature_extraction']
    return generate_caption_helper(
        image,
        decoder,
        encoder,
        tokenizer,
        image_features_extract_model)


# Image Caption Generation query
def image_caption_query(self, instruction,
                        drop=None,
                        epochs=10,
                        preprocess=True,
                        random_state=49,
                        test_size=0.2,
                        top_k=5000,
                        batch_size=32,
                        buffer_size=1000,
                        embedding_dim=256,
                        units=512,
                        gpu=False,
                        generate_plots=True,
                        save_model_decoder=False,
                        save_path_decoder=os.getcwd(),
                        save_model_encoder=False,
                        save_path_encoder=os.getcwd()):

    '''
    function to apply predictive algorithm for image caption generation
    :param many params: used to hyperparametrize the function.
    :return a dictionary object with all of the information for the algorithm.
    '''    
    
    if test_size < 0:
        raise Exception("Test size must be a float between 0 and 1")

    if test_size >= 1:
        raise Exception(
            "Test size must be a float between 0 and 1 (a test size greater than or equal to 1 results in no training "
            "data)")

    if top_k < 1:
        raise Exception("Top_k value must be equal to or greater than 1")

    if batch_size < 1:
        raise Exception("Batch size must be equal to or greater than 1")

    if buffer_size < 1:
        raise Exception("Buffer size must be equal to or greater than 1")

    if embedding_dim < 1:
        raise Exception("Embedding dimension must be equal to or greater than 1")

    if units < 1:
        raise Exception("Units must be equal to or greater than 1")

    if epochs < 1:
        raise Exception("Epoch number is less than 1 (model will not be trained)")

    if save_model_decoder:
        if not os.path.exists(save_path_decoder):
            raise Exception("Decoder save path does not exists")

    if save_model_encoder:
        if not os.path.exists(save_path_encoder):
            raise Exception("Encoder save path does not exists")

    if test_size == 0:
        testing = False
    else:
        testing = True

    if gpu:
        if tf.test.gpu_device_name():
            print('Default GPU Device: {}'.format(tf.test.gpu_device_name()))
        else:
            raise Exception("Please install GPU version of Tensorflow")

        device = '/device:GPU:0'
    else:
        device = '/device:CPU:0'

    np.random.seed(random_state)
    tf.random.set_seed(random_state)

    data = DataReader(self.dataset)
    df = data.data_generator()

    if preprocess:
        df.fillna(0, inplace=True)
    if drop is not None:
        df.drop(drop, axis=1, inplace=True)

    logger("Preprocessing data")

    train_captions = []
    img_name_vector = []
    x = get_path_column(df)
    y = get_similar_column(get_value_instruction(instruction), df)
    logger("->", "Target Column Found: {}".format(y))

    for row in df.iterrows():
        if preprocess:
            caption = '<start> ' + row[1][y] + ' <end>'
        image_id = row[1][x]
        image_path = image_id

        img_name_vector.append(image_path)
        train_captions.append(caption)

    image_model = tf.keras.applications.InceptionV3(include_top=False,
                                                    weights='imagenet')
    new_input = image_model.input
    hidden_layer = image_model.layers[-1].output
    logger("Extracting features from model")
    image_features_extract_model = tf.keras.Model(new_input, hidden_layer)

    image_dataset = tf.data.Dataset.from_tensor_slices(
        sorted(set(img_name_vector)))
    image_dataset = image_dataset.map(
        load_image, num_parallel_calls=tf.data.experimental.AUTOTUNE).batch(16)

    for img, path in image_dataset:
        batch_features = image_features_extract_model(img)
        batch_features = tf.reshape(
            batch_features, (batch_features.shape[0], -1, batch_features.shape[3]))

        for bf, p in zip(batch_features, path):
            path_of_feature = p.numpy().decode("utf-8")
            np.save(path_of_feature, bf.numpy())
    logger("->", "Tokenizing top {} words".format(top_k))
    tokenizer = tf.keras.preprocessing.text.Tokenizer(
        num_words=top_k, oov_token="<unk>", filters='!"#$%&()*+.,-/:;=?@[\]^_`{|}~ ')
    tokenizer.fit_on_texts(train_captions)
    tokenizer.word_index['<pad>'] = 0
    tokenizer.index_word[0] = '<pad>'
    train_seqs = tokenizer.texts_to_sequences(train_captions)
    cap_vector = tf.keras.preprocessing.sequence.pad_sequences(
        train_seqs, padding='post')

    vocab_size = top_k + 1
    num_steps = len(img_name_vector) // batch_size

    if testing:
        img_name_train, img_name_val, cap_train, cap_val = train_test_split(
            img_name_vector, cap_vector, test_size=test_size, random_state=0)
    else:
        img_name_train = img_name_vector
        cap_train = cap_vector

    dataset = tf.data.Dataset.from_tensor_slices((img_name_train, cap_train))

    dataset = dataset.map(lambda item1, item2: tf.numpy_function(
        map_func, [item1, item2], [tf.float32, tf.int32]),
                          num_parallel_calls=tf.data.experimental.AUTOTUNE)

    # Shuffle and batch
    logger("Shuffling dataset")
    dataset = dataset.shuffle(buffer_size).batch(batch_size)
    dataset = dataset.prefetch(buffer_size=tf.data.experimental.AUTOTUNE)

    if testing:
        dataset_val = tf.data.Dataset.from_tensor_slices((img_name_val, cap_val))

        dataset_val = dataset_val.map(lambda item1, item2: tf.numpy_function(
            map_func, [item1, item2], [tf.float32, tf.int32]),
                                      num_parallel_calls=tf.data.experimental.AUTOTUNE)

        # Shuffle and batch
        dataset_val = dataset_val.shuffle(buffer_size).batch(batch_size)
        dataset_val = dataset_val.prefetch(
            buffer_size=tf.data.experimental.AUTOTUNE)

    logger("Establishing encoder decoder framework")
    encoder = CNN_Encoder(embedding_dim)
    decoder = RNN_Decoder(embedding_dim, units, vocab_size)

    optimizer = tf.keras.optimizers.Adam()
    loss_object = tf.keras.losses.SparseCategoricalCrossentropy(
        from_logits=True, reduction='none')

    def loss_function(real, pred):
        mask = tf.math.logical_not(tf.math.equal(real, 0))
        loss_ = loss_object(real, pred)

        mask = tf.cast(mask, dtype=loss_.dtype)
        loss_ *= mask

        return tf.reduce_mean(loss_)

    @tf.function
    def train_step(img_tensor, target):
        with tf.device(device):
            loss = 0

            # initializing the hidden state for each batch
            # because the captions are not related from image to image
            hidden = decoder.reset_state(batch_size=target.shape[0])

            dec_input = tf.expand_dims(
                [tokenizer.word_index['<start>']] * target.shape[0], 1)

            with tf.GradientTape() as tape:
                features = encoder(img_tensor)

                for i in range(1, target.shape[1]):
                    # passing the features through the decoder
                    predictions, hidden, _ = decoder(dec_input, features, hidden)

                    loss += loss_function(target[:, i], predictions)

                    # using teacher forcing
                    dec_input = tf.expand_dims(target[:, i], 1)

            total_loss = (loss / int(target.shape[1]))

            trainable_variables = encoder.trainable_variables + decoder.trainable_variables

            gradients = tape.gradient(loss, trainable_variables)

            optimizer.apply_gradients(zip(gradients, trainable_variables))

            return loss, total_loss

    @tf.function
    def val_step(img_tensor, target):
        with tf.device(device):
            loss = 0

            # initializing the hidden state for each batch
            # because the captions are not related from image to image
            hidden = decoder.reset_state(batch_size=target.shape[0])

            dec_input = tf.expand_dims(
                [tokenizer.word_index['<start>']] * target.shape[0], 1)

            with tf.GradientTape() as tape:
                features = encoder(img_tensor)

                for i in range(1, target.shape[1]):
                    # passing the features through the decoder
                    predictions, hidden, _ = decoder(dec_input, features, hidden)

                    loss += loss_function(target[:, i], predictions)

                    # using teacher forcing
                    dec_input = tf.expand_dims(target[:, i], 1)

            total_loss = (loss / int(target.shape[1]))
            return total_loss

    logger("Training model...")
    with tf.device(device):
        loss_plot_train = []
        loss_plot_val = []
        for epoch in range(epochs):
            total_loss = 0
            total_loss_val = 0

            for (batch, (img_tensor, target)) in enumerate(dataset):
                batch_loss, t_loss = train_step(img_tensor, target)
                total_loss += t_loss

            loss_plot_train.append(total_loss.numpy() / num_steps)

            if testing:
                for (batch, (img_tensor, target)) in enumerate(dataset_val):
                    batch_loss, t_loss = train_step(img_tensor, target)
                    total_loss_val += t_loss

                loss_plot_val.append(total_loss_val.numpy() / num_steps)

    dir_name = os.path.dirname(img_name_vector[0])
    files = os.listdir(dir_name)

    for item in files:
        if item.endswith(".npy"):
            os.remove(os.path.join(dir_name, item))

    plots = {}
    if generate_plots:
        logger("Generating plots")
        plots.update({"loss": libra.plotting.nonkeras_generate_plots.plot_loss(loss_plot_train, loss_plot_val)})

    logger("->", "Final training loss: {}".format(str(total_loss.numpy() / num_steps)))
    total_loss = total_loss.numpy() / num_steps
    if testing:
        total_loss_val = total_loss_val.numpy() / num_steps
        total_loss_val_str = str(total_loss_val)
    else:
        total_loss_val = 0
        total_loss_val_str = str("0, No validation done")

    logger("->", "Final validation loss: {}".format(total_loss_val_str))

    if save_model_decoder:
        logger("Saving decoder checkpoint...")
        encoder.save_weights(save_path_decoder + "decoderImgCap.ckpt")

    if save_model_encoder:
        logger("Saving encoder checkpoint...")
        encoder.save_weights(save_path_encoder + "encoderImgCap.ckpt")

    logger("Storing information in client object under key 'Image Caption'")

    self.models["Image Caption"] = {
        "decoder": decoder,
        "encoder": encoder,
        "tokenizer": tokenizer,
        "feature_extraction": image_features_extract_model,
        "plots": plots,
        'losses': {
            'training_loss': total_loss,
            'validation_loss': total_loss_val
        }
    }
    clearLog()
    return self.models["Image Caption"]



import string
import Levenshtein

# generates mask for sequence to sequence processinng


def produceMask(instruction_label):
    # creates a mask for a string with representation [1, 0, 2] based on
    # character frequencies

    tokenized_string = [char for char in instruction_label]
    ascii_list = list(string.ascii_lowercase)
    res = {i: instruction_label.count(i) for i in set(instruction_label)}

    mask = [0] * len(ascii_list)

    # stores the masks in an object with the base mask and returns
    for x in range(len(ascii_list)):
        if ascii_list[x] in res.keys():
            mask[x] = res[str(ascii_list[x])]

    return mask


def get_similar_column(instruction, dataset):
    # instruction = produceMask(instruction_label)
    distances = []

    for element in dataset.columns:
        value = Levenshtein.distance(instruction, element)
        distances.append(value)

    val, idx = min((val, idx) for (idx, val) in enumerate(distances))

    # returning the column with the index of most similarity
    return dataset.columns[idx]


# exact exame to get_similar_column(). Adapted to allow for small changes
# for model similarity identification
def get_similar_model(model_requested, model_keys):
    distances = []

    for element in model_keys:
        distances.append(Levenshtein.distance(model_requested, element))

    val, idx = min((val, idx) for (idx, val) in enumerate(distances))

    return model_keys[idx]



from libra import client

import unittest

def make_orderer():
    order = {}

    def ordered(f):
        order[f.__name__] = len(order)
        return f

    def compare(a, b):
        return [1, -1][order[a] < order[b]]

    return ordered, compare


ordered, compare = make_orderer()
unittest.defaultTestLoader.sortTestMethodsUsing = compare

class TestQueries(unittest.TestCase):

    newClient = client('tools/data/structured_data/housing.csv')

    """
    TEST QUERIES
    
    Tests some queries in queries.py
    """
    # Tests whether regression_ann_query works without errors, and creates a key in models dictionary
    @ordered
    def test_regression_ann(self):
        self.newClient.regression_query_ann('predict median house value', epochs=3)
        self.assertTrue('regression_ANN' in self.newClient.models)
        del self.newClient.models['regression_ANN']

    # Tests whether classification_ann_query works without errors, and creates a key in models dictionary
    @ordered
    def test_classification_ann(self):
        self.newClient.classification_query_ann('predict ocean proximity', epochs=3)
        self.assertTrue('classification_ANN' in self.newClient.models)
        del self.newClient.models['classification_ANN']

    # Tests whether neural_network_query uses the correct model
    @ordered
    def test_nn_query(self):
        # see if properly chooses regression with a numeric target column
        self.newClient.neural_network_query('predict median house value', epochs=3)
        self.assertTrue('regression_ANN' in self.newClient.models)

        # see if properly chooses classification with a categorical target column
        self.newClient.neural_network_query('predict ocean proximity', epochs=3)
        self.assertTrue('classification_ANN' in self.newClient.models)

    # Tests whether decision_tree_query works without errors, and creates a key in models dictionary
    @ordered
    def test_decision_tree(self):
        self.newClient.decision_tree_query('predict ocean proximity')
        self.assertTrue('decision_tree' in self.newClient.models)

    # Tests whether svm_query works without errors, and creates a key in models dictionary
    @ordered
    def test_svm(self):
        self.newClient.svm_query('predict ocean proximity')
        self.assertTrue('svm' in self.newClient.models)

    # Tests whether nearest_neighbor_query works without errors, and creates a key in models dictionary
    @ordered
    def test_nearest_neighbors(self):
        self.newClient.nearest_neighbor_query('predict ocean proximity')
        self.assertTrue('nearest_neighbor' in self.newClient.models)

    # Tests whether kmeans_clustering_query works without errors, and creates a key in models dictionary
    @ordered
    def test_kmeans(self):
        self.newClient.kmeans_clustering_query('predict ocean proximity')
        self.assertTrue('k_means_clustering' in self.newClient.models)


    """
    TEST ANALYZE() FUNCTION
    
    Tests all branches of .analyze() function in generate_plots
    """
    # Tests analyze() function for k_means_clustering
    @ordered
    def test_analyze_kmeans(self):
        self.newClient.analyze(model='k_means_clustering')
        self.assertTrue('n_centers' in self.newClient.models['k_means_clustering'])
        self.assertTrue('centroids' in self.newClient.models['k_means_clustering'])
        self.assertTrue('inertia' in self.newClient.models['k_means_clustering'])

    # Tests analyze() function on regression_ANN
    @ordered
    def test_analyze_regression(self):
        self.newClient.analyze(model='regression_ANN')
        self.assertTrue('MSE' in self.newClient.models['regression_ANN'])
        self.assertTrue('MAE' in self.newClient.models['regression_ANN'])

    # Tests analyze() function on classification_ANN
    @ordered
    def test_analyze_classification(self):
        self.newClient.analyze(model='classification_ANN')
        self.assertTrue('plots' in self.newClient.models['classification_ANN'])
        self.assertTrue('roc_curve' in self.newClient.models['classification_ANN']['plots'])
        self.assertTrue('confusion_matrix' in self.newClient.models['classification_ANN']['plots'])

        self.assertTrue('scores' in self.newClient.models['classification_ANN'])
        self.assertTrue('recall_score' in self.newClient.models['classification_ANN']['scores'])
        self.assertTrue('precision_score' in self.newClient.models['classification_ANN']['scores'])
        self.assertTrue('f1_score' in self.newClient.models['classification_ANN']['scores'])

    # Tests analyze() function for sklearn models
    @ordered
    def test_analyze_sklearn_classifiers(self):
        for mod in ['svm', 'nearest_neighbor', 'decision_tree']:
            self.newClient.analyze(model=mod)
            modeldict = self.newClient.models[mod]

            self.assertTrue('plots' in modeldict)
            self.assertTrue('roc_curve' in modeldict['plots'])
            self.assertTrue('confusion_matrix' in modeldict['plots'])

            self.assertTrue('scores' in modeldict)
            self.assertTrue('recall_score' in modeldict['scores'])
            self.assertTrue('precision_score' in modeldict['scores'])
            self.assertTrue('f1_score' in modeldict['scores'])

    # Tests invalid model input
    @ordered
    def test_invalid_model(self):
        with self.assertRaises(NameError):
            self.newClient.analyze(model='I dont exist')

if __name__ == '__main__':
    unittest.main()



from colorama import Fore, Style
from tensorflow.keras.callbacks import EarlyStopping
import os
from libra.preprocessing.image_preprocesser import (setwise_preprocessing,
                                                    csv_preprocessing,
                                                    classwise_preprocessing,
                                                    set_distinguisher,
                                                    already_processed)
from libra.preprocessing.data_reader import DataReader
from keras.models import Sequential
from keras.layers import (Dense, Conv2D, Flatten, MaxPooling2D, )
import pandas as pd
from libra.query.supplementaries import save, generate_id
from keras.preprocessing.image import ImageDataGenerator
from sklearn.preprocessing import OneHotEncoder
from libra.plotting.generate_plots import (generate_regression_plots,
                                           generate_classification_plots)
from libra.preprocessing.data_preprocesser import initial_preprocesser
from libra.modeling.prediction_model_creation import get_keras_model_reg, get_keras_model_class
from sklearn.preprocessing import StandardScaler
import numpy as np

os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'

counter = 0
currLog = ""


# # allows for all columns to be displayed when printing()
# pd.options.display.width = None


# # clears the log when new process is started up


def clearLog():
    global counter
    global currLog

    currLog = ""
    counter = 0


# logging function that creates hierarchial display of the processes of
# different functions. Copied into different python files to maintain
# global variable parallels


def logger(instruction, found=""):
    '''
    logging function that creates hierarchial display of the processes of
    different functions. Copied into different python files to maintain
    global variables.

    :param instruction: what you want to be displayed
    :param found: if you want to display something found like target column

    '''
    global counter
    if counter == 0:
        print((" " * 2 * counter) + str(instruction) + str(found))
    elif instruction == "->":
        counter = counter - 1
        print(Fore.BLUE + (" " * 2 * counter) +
              str(instruction) + str(found) + (Style.RESET_ALL))
    else:
        print((" " * 2 * counter) + "|- " + str(instruction) + str(found))
        if instruction == "done...":
            print("\n" + "\n")

    counter += 1


def regression_ann(
        instruction,
        callback=False,
        ca_threshold=None,
        text=[],
        dataset=None,
        drop=None,
        preprocess=True,
        test_size=0.2,
        random_state=49,
        epochs=50,
        generate_plots=True,
        callback_mode='min',
        maximizer="val_loss",
        save_model=False,
        save_path=os.getcwd()):
    '''
    Body of the regression function used that is called in the neural network query
    if the data is numerical.
    :param many parameters: used to preprocess, tune, plot generation, and parameterizing the neural network trained.
    :return dictionary that holds all the information for the finished model.
    '''

    logger("Reading in dataset...")

    dataReader = DataReader(dataset)
    data = dataReader.data_generator()
    # data = pd.read_csv(self.dataset)

    if drop is not None:
        data.drop(drop, axis=1, inplace=True)
    data, y, target, full_pipeline = initial_preprocesser(
        data, instruction, preprocess, ca_threshold, text, test_size=test_size, random_state=random_state)
    logger("->", "Target column found: {}".format(target))

    X_train = data['train']
    X_test = data['test']

    # Target scaling
    target_scaler = StandardScaler()

    y_train = target_scaler.fit_transform(np.array(y['train']).reshape(-1, 1))
    y_test = target_scaler.transform(np.array(y['test']).reshape(-1, 1))

    logger("Establishing callback function")

    models = []
    losses = []
    model_data = []

    # callback function to store lowest loss value
    es = EarlyStopping(
        monitor=maximizer,
        mode=callback_mode,
        verbose=0,
        patience=5)

    callback_value = None
    if callback is not False:
        callback_value = [es]

    i = 0

    # get the first 3 layer model
    model = get_keras_model_reg(data, i)

    logger("Training initial model")
    history = model.fit(
        X_train,
        y_train,
        epochs=epochs,
        validation_data=(
            X_test,
            y_test),
        callbacks=callback_value,
        verbose=0)
    models.append(history)
    model_data.append(model)

    col_name = [["Initial number of layers ",
                 "| Training Loss ", "| Test Loss "]]
    col_width = max(len(word) for row in col_name for word in row) + 2
    for row in col_name:
        print((" " * 2 * counter) + "| " + ("".join(word.ljust(col_width)
                                                    for word in row)) + " |")
    values = []
    values.append(str(len(model.layers)))
    values.append(
        "| " + str(history.history['loss'][len(history.history['val_loss']) - 1]))
    values.append(
        "| " + str(history.history['val_loss'][len(history.history['val_loss']) - 1]))
    datax = []
    datax.append(values)
    for row in datax:
        print((" " * 2 * counter) + "| " + ("".join(word.ljust(col_width)
                                                    for word in row)) + " |")

    losses.append(history.history[maximizer]
                  [len(history.history[maximizer]) - 1])

    # keeps running model and fit functions until the validation loss stops
    # decreasing
    logger("Testing number of layers...")
    col_name = [["Current number of layers", "| Training Loss", "| Test Loss"]]
    col_width = max(len(word) for row in col_name for word in row) + 2
    for row in col_name:
        print((" " * 2 * counter) + "| " + ("".join(word.ljust(col_width)
                                                    for word in row)) + " |")
    datax = []
    while all(x > y for x, y in zip(losses, losses[1:])):
        model = get_keras_model_reg(data, i)
        history = model.fit(
            X_train,
            y_train,
            callbacks=callback_value,
            epochs=epochs,
            validation_data=(
                X_test,
                y_test), verbose=0)
        model_data.append(model)
        models.append(history)

        values = []
        datax = []
        values.append(str(len(model.layers)))
        values.append(
            "| " + str(history.history['loss'][len(history.history['val_loss']) - 1]))
        values.append(
            "| " + str(history.history['val_loss'][len(history.history['val_loss']) - 1]))
        datax.append(values)
        for row in datax:
            print((" " * 2 * counter) + "| " +
                  ("".join(word.ljust(col_width) for word in row)) + " |")
        del values, datax
        losses.append(history.history[maximizer]
                      [len(history.history[maximizer]) - 1])
        i += 1
    # print((" " * 2 * counter)+ tabulate(datax, headers=col_name, tablefmt='orgtbl'))
    final_model = model_data[losses.index(min(losses))]
    final_hist = models[losses.index(min(losses))]
    print("")
    logger('->', "Best number of layers found: " +
           str(len(final_model.layers)))

    logger('->', "Training Loss: " + str(final_hist.history['loss']
                                         [len(final_hist.history['val_loss']) - 1]))
    logger('->', "Test Loss: " + str(final_hist.history['val_loss']
                                     [len(final_hist.history['val_loss']) - 1]))

    # calls function to generate plots in plot generation
    plots = {}
    if generate_plots:
        init_plots, plot_names = generate_regression_plots(
            models[len(models) - 1], data, y)
        for x in range(len(plot_names)):
            plots[str(plot_names[x])] = init_plots[x]

    if save_model:
        save(final_model, save_model)
    # stores values in the client object models dictionary field
    print("")
    logger("Stored model under 'regression_ANN' key")
    clearLog()
    return {
        'id': generate_id(),
        'model': final_model,
        "target": target,
        "plots": plots,
        "preprocesser": full_pipeline,
        "interpreter": target_scaler,
        'test_data': {'X': X_test, 'y': y_test},
        'losses': {
            'training_loss': final_hist.history['loss'],
            'val_loss': final_hist.history['val_loss']}}


def classification_ann(instruction,
                       callback=False,
                       dataset=None,
                       text=[],
                       ca_threshold=None,
                       preprocess=True,
                       callback_mode='min',
                       drop=None,
                       random_state=49,
                       test_size=0.2,
                       epochs=50,
                       generate_plots=True,
                       maximizer="val_accuracy",
                       save_model=False,
                       save_path=os.getcwd()):
    '''
    Body of the classification function used that is called in the neural network query
    if the data is categorical.
    :param many parameters: used to preprocess, tune, plot generation, and parameterizing the neural network trained.
    :return dictionary that holds all the information for the finished model.
    '''
    logger("Reading in dataset...")

    dataReader = DataReader(dataset)
    data = dataReader.data_generator()

    if drop is not None:
        data.drop(drop, axis=1, inplace=True)

    data, y, remove, full_pipeline = initial_preprocesser(
        data, instruction, preprocess, ca_threshold, text, test_size=test_size, random_state=random_state)
    logger("->", "Target column found: {}".format(remove))

    # Needed to make a custom label encoder due to train test split changes
    # Can still be inverse transformed, just a bit of extra work
    y = pd.concat([y['train'], y['test']], axis=0)

    num_classes = len(np.unique(y))

    X_train = data['train']
    X_test = data['test']

    # ANN needs target one hot encoded for classification
    one_hot_encoder = OneHotEncoder()

    y = pd.DataFrame(
        one_hot_encoder.fit_transform(
            np.reshape(
                y.values,
                (-1,
                 1))).toarray(),
        columns=one_hot_encoder.get_feature_names())

    y_train = y.iloc[:len(X_train)]
    y_test = y.iloc[len(X_train):]

    models = []
    losses = []
    accuracies = []
    model_data = []

    logger("Establishing callback function...")

    # early stopping callback
    es = EarlyStopping(
        monitor=maximizer,
        mode='max',
        verbose=0,
        patience=5)

    callback_value = None
    if callback is not False:
        callback_value = [es]

    i = 0
    model = get_keras_model_class(data, i, num_classes)
    logger("Training initial model...")

    history = model.fit(
        X_train,
        y_train,
        callbacks=callback_value,
        epochs=epochs,
        validation_data=(
            X_test,
            y_test),
        verbose=0)

    model_data.append(model)
    models.append(history)
    col_name = [["Initial number of layers ",
                 "| Training Accuracy ", "| Test Accuracy "]]
    col_width = max(len(word) for row in col_name for word in row) + 2
    for row in col_name:
        print((" " * 2 * counter) + "| " + ("".join(word.ljust(col_width)
                                                    for word in row)) + " |")
    values = []
    values.append(str(len(model.layers)))
    values.append(
        "| " + str(history.history['accuracy'][len(history.history['val_accuracy']) - 1]))
    values.append(
        "| " + str(history.history['val_accuracy'][len(history.history['val_accuracy']) - 1]))
    datax = []
    datax.append(values)
    for row in datax:
        print((" " * 2 * counter) + "| " + ("".join(word.ljust(col_width)
                                                    for word in row)) + " |")
    # print((" " * 2 * counter)+ tabulate(datax, headers=col_name, tablefmt='orgtbl'))
    losses.append(history.history[maximizer]
                  [len(history.history[maximizer]) - 1])
    # keeps running model and fit functions until the validation loss stops
    # decreasing

    logger("Testing number of layers...")
    col_name = [["Current number of layers",
                 "| Training Accuracy", "| Test Accuracy"]]
    col_width = max(len(word) for row in col_name for word in row) + 2

    for row in col_name:
        print((" " * 2 * counter) + "| " + ("".join(word.ljust(col_width)
                                                    for word in row)) + " |")
    datax = []
    while all(x < y for x, y in zip(accuracies, accuracies[1:])):
        model = get_keras_model_class(data, i, num_classes)
        history = model.fit(
            X_train,
            y_train,
            callbacks=callback_value,
            epochs=epochs,
            validation_data=(
                X_test,
                y_test),
            verbose=0)

        values = []
        datax = []
        values.append(str(len(model.layers)))
        values.append(
            "| " + str(history.history['accuracy'][len(history.history['val_accuracy']) - 1]))
        values.append(
            "| " + str(history.history['accuracy'][len(history.history['val_accuracy']) - 1]))
        datax.append(values)
        for row in datax:
            print((" " * 2 * counter) + "| " +
                  ("".join(word.ljust(col_width) for word in row)) + " |")
        del values, datax
        losses.append(history.history[maximizer]
                      [len(history.history[maximizer]) - 1])
        accuracies.append(history.history['val_accuracy']
                          [len(history.history['val_accuracy']) - 1])
        models.append(history)
        model_data.append(model)

        i += 1
    # print((" " * 2 * counter)+ tabulate(datax, headers=col_name, tablefmt='orgtbl'))
    # del values, datax

    final_model = model_data[accuracies.index(max(accuracies))]
    final_hist = models[accuracies.index(max(accuracies))]

    print("")
    logger('->', "Best number of layers found: " +
           str(len(final_model.layers)))
    logger('->', "Training Accuracy: " + str(final_hist.history['accuracy']
                                             [len(final_hist.history['val_accuracy']) - 1]))
    logger('->', "Test Accuracy: " + str(final_hist.history['val_accuracy'][
        len(final_hist.history['val_accuracy']) - 1]))

    # genreates appropriate classification plots by feeding all information
    plots = {}
    if generate_plots:
        plots = generate_classification_plots(
            models[len(models) - 1], data, y, model, X_test, y_test)

    if save_model:
        save(final_model, save_model)

    print("")
    logger("Stored model under 'classification_ANN' key")
    clearLog()
    # stores the values and plots into the object dictionary
    return {
        'id': generate_id(),
        "model": final_model,
        'num_classes': num_classes,
        "plots": plots,
        "target": remove,
        "preprocesser": full_pipeline,
        "interpreter": one_hot_encoder,
        'test_data': {'X': X_test, 'y': y_test},
        'losses': {
            'training_loss': final_hist.history['loss'],
            'val_loss': final_hist.history['val_loss']},
        'accuracy': {
            'training_accuracy': final_hist.history['accuracy'],
            'validation_accuracy': final_hist.history['val_accuracy']}}


def convolutional(instruction=None,
                  read_mode=None,
                  preprocess=True,
                  verbose=0,
                  data_path=os.getcwd(),
                  new_folders=True,
                  image_column=None,
                  training_ratio=0.8,
                  augmentation=True,
                  epochs=10,
                  height=None,
                  width=None):
    '''
    Body of the convolutional function used that is called in the neural network query
    if the data is presented in images.
    :param many parameters: used to preprocess, tune, plot generation, and parameterizing the convolutional neural network trained.
    :return dictionary that holds all the information for the finished model.
    '''

    logger("Generating datasets for classes...")

    if preprocess:
        read_mode_info = set_distinguisher(data_path, read_mode)
        read_mode = read_mode_info["read_mode"]

        training_path = "/proc_training_set"
        testing_path = "/proc_testing_set"

        if read_mode == "setwise":
            processInfo = setwise_preprocessing(
                data_path, new_folders, height, width)
            if not new_folders:
                training_path = "/training_set"
                testing_path = "/testing_set"

        # if image dataset in form of csv
        elif read_mode == "csvwise":
            if training_ratio <= 0 or training_ratio >= 1:
                raise BaseException(f"Test ratio must be between 0 and 1.")
            processInfo = csv_preprocessing(read_mode_info["csv_path"],
                                            data_path,
                                            instruction,
                                            image_column,
                                            training_ratio,
                                            height,
                                            width)

        # if image dataset in form of one folder containing class folders
        elif read_mode == "classwise":
            if training_ratio <= 0 or training_ratio >= 1:
                raise BaseException(f"Test ratio must be between 0 and 1.")
            processInfo = classwise_preprocessing(
                data_path, training_ratio, height, width)

    else:
        training_path = "/training_set"
        testing_path = "/testing_set"
        processInfo = already_processed(data_path)

    input_shape = (processInfo["height"], processInfo["width"], 3)
    input_single = (processInfo["height"], processInfo["width"])
    num_classes = processInfo["num_categories"]
    loss_func = ""

    if num_classes > 2:
        loss_func = "categorical_crossentropy"
    elif num_classes == 2:
        loss_func = "binary_crossentropy"

    logger("Creating convolutional neural network dynamically...")
    # Convolutional Neural Network
    model = Sequential()
    model.add(
        Conv2D(
            64,
            kernel_size=3,
            activation="relu",
            input_shape=input_shape))
    model.add(MaxPooling2D(pool_size=(2, 2)))
    model.add(Conv2D(64, kernel_size=3, activation="relu"))
    model.add(MaxPooling2D(pool_size=(2, 2)))
    model.add(Flatten())
    model.add(Dense(num_classes, activation="softmax"))
    model.compile(
        optimizer="adam",
        loss=loss_func,
        metrics=['accuracy'])
    logger("Located image data")
    if augmentation:
        train_data = ImageDataGenerator(rescale=1. / 255,
                                        shear_range=0.2,
                                        zoom_range=0.2,
                                        horizontal_flip=True)
        test_data = ImageDataGenerator(rescale=1. / 255)

        logger('Dataset augmented through zoom, shear, flip, and rescale')
    else:
        train_data = ImageDataGenerator()
        test_data = ImageDataGenerator()

    logger("->", "Optimal image size identified: {}".format(input_shape))
    X_train = train_data.flow_from_directory(data_path + training_path,
                                             target_size=input_single,
                                             color_mode='rgb',
                                             batch_size=(32 if processInfo["train_size"] >= 32 else 1),
                                             class_mode=loss_func[:loss_func.find("_")])
    X_test = test_data.flow_from_directory(data_path + testing_path,
                                           target_size=input_single,
                                           color_mode='rgb',
                                           batch_size=(32 if processInfo["test_size"] >= 32 else 1),
                                           class_mode=loss_func[:loss_func.find("_")])


    if epochs < 0:
        raise BaseException("Number of epochs has to be greater than 0.")
    logger('Training image model')
    history = model.fit(
        X_train,
        steps_per_epoch=X_train.n //
        X_train.batch_size,
        validation_data=X_test,
        validation_steps=X_test.n //
        X_test.batch_size,
        epochs=epochs,
        verbose=verbose)

    logger('->', 'Final training accuracy: {}'.format(history.history['accuracy'][len(history.history['accuracy']) - 1]))
    logger('->', 'Final validation accuracy: {}'.format(history.history['val_accuracy'][len(history.history['val_accuracy']) - 1]))
    # storing values the model dictionary

    logger("Stored model under 'convolutional_NN' key")
    clearLog()
    return {
        'id': generate_id(),
        'data_type': read_mode,
        'data_path': data_path,
        'data': {'train': X_train, 'test': X_test},
        'shape': input_shape,
        "model": model,
        'losses': {
            'training_loss': history.history['loss'],
            'val_loss': history.history['val_loss']},
        'accuracy': {
            'training_accuracy': history.history['accuracy'],
            'validation_accuracy': history.history['val_accuracy']},
        'num_classes': (2 if num_classes == 1 else num_classes),
        'data_sizes': {'train_size': processInfo['train_size'], 'test_size': processInfo['test_size']}}




import os
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'

from kerastuner.applications import HyperResNet
from tensorflow.keras.layers import (Conv2D,
                                     MaxPooling2D,
                                     Dense,
                                     Flatten,
                                     Dropout)
from kerastuner import HyperModel
from kerastuner.tuners import RandomSearch, Hyperband
from sklearn import preprocessing
from tensorflow import keras
from sklearn.model_selection import train_test_split
import tensorflow as tf
from sklearn.preprocessing import StandardScaler


tf.compat.v1.logging.set_verbosity(tf.compat.v1.logging.ERROR)

# creates hypermodel class for CNN tuning

class CNNHyperModel(HyperModel):
    def __init__(self, input_shape, num_classes):
        self.input_shape = input_shape
        self.num_classes = num_classes

    # model builder: creates different layers with a selection of parameters
    # that during runtime it selects to use
    def build(self, hp):
        model = keras.Sequential()
        model.add(
            Conv2D(
                filters=16,
                kernel_size=3,
                activation='relu',
                input_shape=self.input_shape
            )
        )
        model.add(
            Conv2D(
                filters=16,
                activation='relu',
                kernel_size=3
            )
        )
        model.add(MaxPooling2D(pool_size=2))
        model.add(
            Dropout(rate=hp.Float(
                'dropout_1',
                min_value=0.0,
                max_value=0.5,
                default=0.25,
                step=0.05,
            ))
        )
        model.add(
            Conv2D(
                filters=32,
                kernel_size=3,
                activation='relu'
            )
        )
        model.add(
            Conv2D(
                filters=16,
                activation='relu',
                kernel_size=3
            )
        )
        model.add(MaxPooling2D(pool_size=2))
        model.add(
            Dropout(rate=hp.Float(
                'dropout_2',
                min_value=0.0,
                max_value=0.5,
                default=0.25,
                step=0.05,
            ))
        )
        model.add(Flatten())
        model.add(
            Dense(
                units=hp.Int(
                    'units',
                    min_value=32,
                    max_value=512,
                    step=32,
                    default=128
                ),
                activation=hp.Choice(
                    'dense_activation',
                    values=['relu', 'tanh', 'sigmoid'],
                    default='relu'
                )
            )
        )
        model.add(
            Dropout(
                rate=hp.Float(
                    'dropout_3',
                    min_value=0.0,
                    max_value=0.5,
                    default=0.25,
                    step=0.05
                )
            )
        )
        model.add(Dense(self.num_classes, activation='softmax'))

        model.compile(
            optimizer=keras.optimizers.Adam(
                hp.Float(
                    'learning_rate',
                    min_value=1e-4,
                    max_value=1e-2,
                    sampling='LOG',
                    default=1e-3)),
            loss=(
                'binary_crossentropy' if self.num_classes == 2 else 'categorical_crossentropy'),
            metrics=['accuracy'])
        return model


def tuneReg(
        data,
        target,
        max_layers=10,
        min_layers=2,
        min_dense=32,
        max_dense=512,
        executions_per_trial=1,
        max_trials=5,
        epochs=10,
        activation='relu',
        directory='my_dir',
        step=32,
        verbose=0,
        test_size=0.2
):
    # function build model using hyperparameter
    def build_model(hp):
        model = keras.Sequential()
        for i in range(hp.Int('num_layers', min_layers, max_layers)):
            model.add(Dense(units=hp.Int('units_' + str(i),
                                         min_value=min_dense,
                                         max_value=max_dense,
                                         step=step),
                            activation=activation))
            model.add(Dropout(rate=hp.Float(
                              'dropout_3',
                              min_value=0.0,
                              max_value=0.5,
                              default=0.20,
                              step=0.05)))
        model.add(Dense(1, activation='linear'))
        model.compile(
            optimizer=keras.optimizers.Adam(
                                       hp.Float('learning_rate',
                                                min_value=1e-5,
                                                max_value=1e-2,
                                                sampling='LOG',
                                                default=1e-3)),
            loss='mse',
            metrics=['accuracy'])
        return model

    # random search for the model
    tuner = RandomSearch(
        build_model,
        objective='loss',
        max_trials=max_trials,
        executions_per_trial=executions_per_trial,
        directory=directory)
    # tuner.search_space_summary()
    # del data[target]

    X_train, X_test, y_train, y_test = train_test_split(
        data, target, test_size=0.2, random_state=49)

    # searches the tuner space defined by hyperparameters (hp) and returns the
    # best model

    tuner.search(X_train, y_train,
                 epochs=epochs,
                 validation_data=(X_test, y_test),
                 callbacks=[tf.keras.callbacks.TensorBoard('my_dir')])

    models = tuner.get_best_models(num_models=1)[0]
    hyp = tuner.get_best_hyperparameters(num_trials=1)[0]
    history = tuner_hist(
        data,
        target,
        tuner,
        hyp,
        epochs=epochs,
        verbose=verbose,
        test_size=test_size)
    """
    Return:
        models[0] : best model obtained after tuning
        best_hps : best Hyperprameters obtained after tuning, stored as map
        history : history of the data executed from the given model
    """
    return models, hyp, history, X_test, y_test


def tuneClass(
        X,
        y,
        num_classes,
        max_layers=10,
        min_layers=2,
        min_dense=32,
        max_dense=512,
        executions_per_trial=1,
        max_trials=3,
        activation='relu',
        directory='my_dir',
        loss='categorical_crossentropy',
        metrics='accuracy',
        epochs=10,
        step=32,
        verbose=0,
        test_size=0.2):
    # function build model using hyperparameter
    le = preprocessing.LabelEncoder()
    y = tf.keras.utils.to_categorical(
        le.fit_transform(y), num_classes=num_classes)

    def build_model(hp):
        model = keras.Sequential()
        for i in range(hp.Int('num_layers', min_layers, max_layers)):
            model.add(Dense(units=hp.Int('units_' + str(i),
                                         min_value=min_dense,
                                         max_value=max_dense,
                                         step=step),
                            activation=activation))
        model.add(Dense(num_classes, activation='softmax'))
        model.compile(
            optimizer=keras.optimizers.Adam(
                                       hp.Float('learning_rate',
                                                min_value=1e-5,
                                                max_value=1e-2,
                                                sampling='LOG',
                                                default=1e-3)),
            loss=loss,
            metrics=[metrics])
        return model
    
    # tuners, establish the object to look through the tuner search space
    tuner = RandomSearch(
        build_model,
        objective='val_accuracy',
        max_trials=max_trials,
        executions_per_trial=executions_per_trial,
        directory=directory,
        project_name='class_tuned')

    # tuner.search_space_summary()
    X_train, X_test, y_train, y_test = train_test_split(
        X, y, test_size=0.2, random_state=49)

    # searches the tuner space defined by hyperparameters (hp) and returns the
    # best model
    tuner.search(X_train, y_train,
                 epochs=epochs,
                 validation_data=(X_test, y_test))
    models = tuner.get_best_models(num_models=1)[0]
    hyp = tuner.get_best_hyperparameters(num_trials=1)[0]
    #hyp = tuner.oracle.get_best_trials(num_trials=1)[0].hyperparameters.values
    #best_hps = np.stack(hyp).astype(None)
    history = tuner_hist(
        X,
        y,
        tuner,
        hyp,
        epochs=epochs,
        verbose=verbose,
        test_size=test_size)

    """
    Return:
        models[0] : best model obtained after tuning
        best_hps : best Hyperprameters obtained after tuning, stored as array
        history : history of the data executed from the given model
    """
    return models, hyp, history, X_test, y_test


def tuneCNN(
        X_train,
        X_test,
        height,
        width,
        num_classes,
        patience=1,
        executions_per_trial=1,
        seed=42,
        max_trials=3,
        objective='val_accuracy',
        directory='my_dir',
        epochs=10,
        verbose=0,
        test_size=0.2):
    # creates hypermodel object based on the num_classes and the input shape
    hypermodel = CNNHyperModel(input_shape=(
        height, width, 3), num_classes=num_classes)

    # # tuners, establish the object to look through the tuner search space
    tuner = RandomSearch(
        hypermodel,
        objective=objective,
        seed=seed,
        max_trials=max_trials,
        executions_per_trial=executions_per_trial,
        directory=directory,
    )


    # searches the tuner space defined by hyperparameters (hp) and returns the
    # best model

    tuner.search(X_train,
                 validation_data=X_test,
                 callbacks=[tf.keras.callbacks.EarlyStopping(patience=patience)],
                 epochs=epochs,
                 verbose=verbose)

    # best hyperparamters
    hyp = tuner.get_best_hyperparameters(num_trials=1)[0]
    #hyp = tuner.oracle.get_best_trials(num_trials=1)[0].hyperparameters.values
    #best_hps = np.stack(hyp).astype(None)
    history = tuner_hist(
        X_train,
        X_test,
        tuner,
        hyp,
        img=1,
        epochs=epochs,
        verbose=verbose,
        test_size=test_size)

    """
    Return:
        models[0] : best model obtained after tuning
        best_hps : best Hyperprameters obtained after tuning, stored as array
        history : history of the data executed from the given model
    """
    return tuner.get_best_models(1)[0], hyp, history


def tuneHyperband(X,
                  y,
                  max_trials=3):
    """
    Perform Hyperband Tuning to search for the best model and Hyperparameters
    Arguments:
        X: Input dataset
        y: Label or output dataset
        max_trials: Trials required to perform tuning
    """
    hypermodel = HyperResNet(input_shape=(128, 128, 3), num_classes=10)
    tuner = Hyperband(
        hypermodel,
        max_epochs=max_trials,
        objective='val_accuracy',
        seed=42,
    )

    X_train, X_test, y_train, y_test = train_test_split(
        X, y, test_size=0.2, random_state=49)

    # searches the tuner space defined by hyperparameters (hp) and returns the
    # best model
    tuner.search(X_train, y_train,
                 epochs=5,
                 validation_data=(X_test, y_test))
    hyp = tuner.get_best_hyperparameters(num_trials=1)[0]
    #hyp = tuner.oracle.get_best_trials(num_trials=1)[0].hyperparameters.values
    #best_hps = np.stack(hyp).astype(None)

    history = tuner_hist(X, y, tuner, hyp)
    """
    Return:
        models[0] : best model obtained after tuning
        best_hps : best Hyperprameters obtained after tuning, stored as array
        history : history of the data executed from the given model
    """
    return tuner.get_best_models(1)[0], hyp, history


def tuner_hist(
        X,
        y,
        tuner,
        best_hps,
        img=0,
        epochs=5,
        test_size=0.2,
        verbose=0):
    model = tuner.hypermodel.build(best_hps)

    if img == 0:
        X_train, X_test, y_train, y_test = train_test_split(
            X, y, test_size=test_size, random_state=49)
        history = model.fit(X_train, y_train,
                            epochs=epochs,
                            validation_data=(X_test, y_test),
                            verbose=verbose)
    else:
        history = model.fit_generator(X,
                                      epochs=epochs,
                                      validation_data=y,
                                      verbose=verbose)

    return history



from sklearn.feature_selection import SelectFromModel
from sklearn import preprocessing, tree
from itertools import product, permutations
from libra.preprocessing.data_reader import DataReader
import os
from sklearn.ensemble import RandomForestRegressor
from libra.preprocessing.data_preprocesser import structured_preprocesser, initial_preprocesser
from libra.data_generation.grammartree import get_value_instruction
from libra.data_generation.dataset_labelmatcher import get_similar_column

from xgboost import XGBClassifier

from sklearn.model_selection import train_test_split
from sklearn import preprocessing
from sklearn.metrics import accuracy_score
from sklearn.decomposition import PCA, FastICA, KernelPCA
import pandas as pd
import numpy as np
from colorama import Fore, Style
# function imports from other files

counter = 0

# allows for all columns to be displayed when printing()
pd.options.display.width = None


def logger(instruction, found=""):
    global counter
    if counter == 0:
        print((" " * 2 * counter) + str(instruction) + str(found))
    elif instruction == "->":
        counter = counter - 1
        print(Fore.BLUE + (" " * 2 * counter) +
              str(instruction) + str(found) + (Style.RESET_ALL))
    else:
        print((" " * 2 * counter) + "|- " + str(instruction) + str(found))
        if instruction == "done...":
            print("\n" + "\n")

    counter += 1


def printtable(col_name, col_width):
    global counter
    for row in col_name:
        print((" " * 2 * counter) + "| " + ("".join(word.ljust(col_width)
                                                    for word in row)) + " |")


def dimensionality_reduc(
        instruction,
        dataset,
        arr=[
            "RF",
            "PCA",
            "KPCA",
            "ICA"],
        inplace=False):
    '''
    function to perform dimensionality reduction on the dataset (retrieve only 
    features with most relevance from multidimensional space of the dataset)
    :param instruction: command sent to client instance in written query
    :param dataset: data instantiated in client instance passed to the algorithm
    :param arr: list of options of algorithm/dimension reducing techniques 
    options to choose from
    :param inplace: option to keep features that were deemed as not important
    intact in the dataset
    '''
    
    global counter
    
    dataReader = DataReader(dataset)

    logger("loading dataset...")
    data = dataReader.data_generator()
    data.fillna(0, inplace=True)

    logger("getting most similar column from instruction...")
    target = get_similar_column(get_value_instruction(instruction), data)

    y = data[target]
    del data[target]
    le = preprocessing.LabelEncoder()
    y = le.fit_transform(y)

    data = structured_preprocesser(data)

    perms = []
    overall_storage = []
    finals = []

    logger("generating dimensionality permutations...")
    for i in range(1, len(arr) + 1):
        for elem in list(permutations(arr, i)):
            perms.append(elem)

    logger("running each possible permutation...")
    logger("realigning tensors...")
    for path in perms:
        currSet = data
        for element in path:
            if element == "RF":
                data_mod, beg_acc, final_acc, col_removed = dimensionality_RF(
                    instruction, currSet, target, y)
            elif element == "PCA":
                data_mod, beg_acc, final_acc, col_removed = dimensionality_PCA(
                    instruction, currSet, target, y)
            elif element == "KPCA":
                data_mod, beg_acc, final_acc, col_removed = dimensionality_KPCA(
                    instruction, currSet, target, y)
            elif element == "ICA":
                data_mod, beg_acc, final_acc, col_removed = dimensionality_ICA(
                    instruction, currSet, target, y)
            overall_storage.append(
                list([data_mod, beg_acc, final_acc, col_removed]))
            currSet = data_mod
        finals.append(overall_storage[len(overall_storage) - 1])

    logger("Fetching Best Accuracies...")
    accs = []
    logger("->", "Baseline Accuracy: " + str(finals[0][1]))
    # print("----------------------------")
    col_name = [["Permutation ", "| Final Accuracy "]]
    printtable(col_name, max(len(word)
                             for row in col_name for word in row) + 5)
    for i, element in product(range(len(finals)), finals):
        values = []
        values.append(str(perms[i]))
        values.append("| " + str(element[2]))
        datax = []
        datax.append(values)
        printtable(datax, max(len(word)
                              for row in col_name for word in row) + 5)
        del values, datax
        if finals[0][1] < element[2]:
            accs.append(list([str(perms[i]),
                              "| " + str(element[2])]))
    print("")
    logger("->", " Best Accuracies")
    # print("----------------------------")
    col_name = [["Permutation ", "| Final Accuracy "]]
    printtable(col_name, max(len(word)
                             for row in col_name for word in row) + 5)
    printtable(accs, col_width)

    if inplace:
        data.to_csv(dataset)


def dimensionality_RF(instruction, dataset, target="", y="", n_features=10):
    '''
    function to reduce dimensionality in dataset via random forest method
    :param instruction: command sent to client instance in written query.
    :param dataset: data instantiated in client instance passed to the algorithm
    :param target: column name of response variable/feature
    :param y: dictionary of train/test data values associated with response variable/feature
    :param n_features: maximum number of features to choose to analyze/select
    '''
    
    global counter

    dataReader = DataReader("./data/" + get_last_file()[0])

    if target == "":
        data = dataReader.data_generator()
        data.fillna(0, inplace=True)
        remove = get_similar_column(get_value_instruction(instruction), data)
        data = structured_preprocesser(data)

        y = data[remove]
        del data[remove]
        le = preprocessing.LabelEncoder()
        y = le.fit_transform(y)

    X_train, X_test, y_train, y_test = train_test_split(
        dataset, y, test_size=0.2, random_state=49)
    first_classifier = tree.DecisionTreeClassifier()
    first_classifier.fit(X_train, y_train)

    first_classifier_acc = accuracy_score(
        first_classifier.predict(X_test), y_test)

    accuracy_scores = [first_classifier_acc]
    columns = []
    datas = []
    datas.append(dataset)
    columns.append([])

    for i, x in product(range(3, 10), range(4, len(dataset.columns))):
        feature_model = RandomForestRegressor(random_state=1, max_depth=i)
        feature_model.fit(X_train, y_train)

        importances = feature_model.feature_importances_
        indices = np.argsort(importances)[-x:]
        columns.append(dataset.columns[indices])

        X_temp_train = X_train[dataset.columns[indices]]
        X_temp_test = X_test[dataset.columns[indices]]

        val = pd.DataFrame(np.r_[X_temp_train, X_temp_test])
        val[target] = np.r_[y_train, y_test]
        datas.append(val)

        vr = tree.DecisionTreeClassifier()
        vr.fit(X_temp_train, y_train)

        accuracy_scores.append(accuracy_score(vr.predict(X_temp_test), y_test))

    the_index = accuracy_scores.index(max(accuracy_scores))

    return datas[the_index], accuracy_scores[0], max(
        accuracy_scores), list(columns[the_index])


def dimensionality_PCA(instruction, dataset, ca_threshold=None):
     '''
     function to reduce dimensionality in dataset via principal component analysis method
     :param instruction: command sent to client instance in written query.
     :param dataset: data instantiated in client instance passed to the algorithm
     :param ca_threshold: percentage of dataset to be preprocessed using morphological component analysis
     '''
        
    global counter

    pca = PCA(0.92)

    data, y, target, full_pipeline = initial_preprocesser(
        dataset, instruction, ca_threshold=ca_threshold, preprocess=True)

    X_train = data['train']
    X_test = data['test']

    y_train = y['train']
    y_test = y['test']

    X_train_mod = pca.fit_transform(X_train)
    X_test_mod = pca.transform(X_test)

    clf = tree.DecisionTreeClassifier()
    clf_mod = tree.DecisionTreeClassifier()

    clf.fit(X_train, y_train)
    clf_mod.fit(X_train_mod, y_train)

    acc = []
    acc.append(accuracy_score(
        clf_mod.predict(X_test_mod), y_test))
    for i, j in product(range(3, 10), ["entropy", "gini"]):
        model = tree.DecisionTreeClassifier(criterion=j, max_depth=i)
        model = model.fit(X_train_mod, y_train)
        acc.append(accuracy_score(model.predict(X_test_mod), y_test))
    del i, j

    data_modified = pd.concat(
        [pd.DataFrame(X_train_mod), pd.DataFrame(X_test_mod)], axis=0)

    y_combined = np.r_[y_train, y_test]
    data_modified[target] = y_combined
    # data_modified.to_csv("./data/housingPCA.csv")

    return data_modified, accuracy_score(
        clf.predict(X_test), y_test), max(acc), (len(
            dataset.columns) - len(data_modified.columns))


def dimensionality_ICA(instruction, dataset, target="", y=""):
    '''
    function to reduce dimensionality in dataset via independent component analysis
    :param instruction: command sent to client instance in written query.
    :param dataset: data instantiated in client instance passed to the algorithm
    :param target: column name of response variable/feature
    :param y: dictionary of train/test data values associated with response variable/feature
    '''
    global counter

    dataReader = DataReader("./data/" + get_last_file()[0])

    if target == "":
        data = dataReader.data_generator()
        data.fillna(0, inplace=True)
        remove = get_similar_column(get_value_instruction(instruction), data)

        y = data[remove]
        del data[remove]
        le = preprocessing.LabelEncoder()
        y = le.fit_transform(y)

    pca = FastICA(n_components=len(dataset.columns))
    data_modified = pca.fit_transform(dataset)

    X_train, X_test, y_train, y_test = train_test_split(
        dataset, y, test_size=0.2, random_state=49)
    X_train_mod, X_test_mod, y_train_mod, y_test_mod = train_test_split(
        data_modified, y, test_size=0.2, random_state=49)

    clf = tree.DecisionTreeClassifier()
    clf.fit(X_train, y_train)

    clf_mod = tree.DecisionTreeClassifier()
    clf_mod.fit(X_train_mod, y_train_mod)
    acc = []
    acc.append(accuracy_score(
        clf_mod.predict(X_test_mod), y_test_mod))
    for i, j in product(range(3, 10), ["entropy", "gini"]):
        model = tree.DecisionTreeClassifier(criterion=j, max_depth=i)
        model = model.fit(X_train, y_train)
        acc.append(accuracy_score(model.predict(X_test), y_test))
    del i, j
    data_modified = pd.DataFrame(data_modified)
    data_modified[target] = np.r_[y_train, y_test]
    # data_modified.to_csv("./data/housingPCA.csv")

    return data_modified, accuracy_score(
        clf.predict(X_test), y_test), max(acc), (len(
            dataset.columns) - len(data_modified.columns))


def get_last_file():
    '''
    function to retrieve most recently created file generated from
    reduced dimensionality technique
    '''
    
    max_mtime = 0
    for dirname, subdirs, files in os.walk("./data"):
        for fname in files:
            if fname == ".DS_Store":
                continue
            full_path = os.path.join(dirname, fname)
            mtime = os.stat(full_path).st_mtime
            if mtime > max_mtime:
                max_mtime = mtime
                max_dir = dirname
                max_file = fname
    return max_file, max_dir, max_mtime


def dimensionality_KPCA(instruction, dataset, target="", y=""):
    '''
    function to reduce dimensionality in dataset via kernal principal component analysis
    :param instruction: command sent to client instance in written query.
    :param dataset: data instantiated in client instance passed to the algorithm
    :param target: column name of response variable/feature
    :param y: dictionary of train/test data values associated with response variable/feature
    '''
    
    global counter

    dataReader = DataReader("./data/" + get_last_file()[0])

    if target == "":
        data = dataReader.data_generator()
        data.fillna(0, inplace=True)
        remove = get_similar_column(get_value_instruction(instruction), data)

        y = data[remove]
        del data[remove]
        le = preprocessing.LabelEncoder()
        y = le.fit_transform(y)

    kpca = KernelPCA(n_components=len(dataset.columns), kernel="rbf")
    data_modified = kpca.fit_transform(dataset)

    X_train, X_test, y_train, y_test = train_test_split(
        dataset, y, test_size=0.2, random_state=49)
    X_train_mod, X_test_mod, y_train_mod, y_test_mod = train_test_split(
        data_modified, y, test_size=0.2, random_state=49)

    clf = tree.DecisionTreeClassifier()
    clf.fit(X_train, y_train)

    clf_mod = tree.DecisionTreeClassifier()
    clf_mod.fit(X_train_mod, y_train_mod)
    acc = []
    acc.append(accuracy_score(
        clf_mod.predict(X_test_mod), y_test_mod))
    for i, j in product(range(3, 10), ["entropy", "gini"]):
        model = tree.DecisionTreeClassifier(criterion=j, max_depth=i)
        model = model.fit(X_train_mod, y_train_mod)
        acc.append(accuracy_score(model.predict(X_test_mod), y_test))
    del i, j
    data_modified = pd.DataFrame(data_modified)
    data_modified[target] = np.r_[y_train, y_test]
    # data_modified.to_csv("./data/housingPCA.csv")

    return data_modified, accuracy_score(
        clf.predict(X_test), y_test), max(acc), (len(
            dataset.columns) - len(data_modified.columns))

    def booster(dataset, obj):
        #obj=["reg:linear","multi:softmax "]

        X_train, X_test, y_train, y_test = train_test_split(
            dataset, y, test_size=0.2, random_state=49)
        clf = XGBClassifier(
            objective=obj,
            learning_rate=0.1,
            silent=1,
            alpha=10)
        clf.fit(X_train, y_train)
        return accuracy_score(clf.predict(X_test_mod), y_test_mod)
        # importance graph
        #plt.rcParams['figure.figsize'] = [5, 5]
        # plt.show()


#dimensionalityPCA("Predict median house value", "./data/housing.csv")

#dimensionalityReduc("Predict ocean_proximity", "./data/housing.csv")

# data = pd.read_csv("./data/housing.csv")
#
# print(dimensionality_PCA("Model ocean proximity", data))



from selenium import webdriver
from selenium.webdriver.common.keys import Keys
import json
import os
import urllib
import argparse
import requests


# def generate_data(word):
#     # identifies the search term and the base google address
#     searchterm = word
#     url = "https://www.google.co.in/search?q=" + \
#         searchterm + "&source=lnms&tbm=isch"

#     # identifies the chrome driver and the location to store the images
#     PROJECT_ROOT = os.path.abspath(os.path.dirname(__file__))
#     DRIVER_BIN = os.path.join(PROJECT_ROOT, "chromedriver")

#     # establishes web browser and launches
#     browser = webdriver.Chrome(executable_path=DRIVER_BIN)
#     browser.get(url)
#     img_count = 0

#     # extensions for images that're accepted
#     extensions = {"jpg", "jpeg", "png", "gif"}
#     if not os.path.exists(searchterm):
#         os.mkdir(searchterm)

#     # continues to scroll down the page when it reaches the bottom
#     for _ in range(500):
#         browser.execute_script("window.scrollBy(0,10000)")

#     html = browser.page_source.split('["')
#     imges = []

#     # identifies the url of the image, if it exists, and processes it
#     # appropriately
#     for i in html:
#         if i.startswith('http') and i.split(
#                 '"')[0].split('.')[-1] in extensions:
#             imges.append(i.split('"')[0])

#     i = 0

#     # splits into only neccesary part to identify image and download
#     for link in imges:
#         filename = link.split('/')[-1]
#         urllib.urlretrieve(link, searchterm + "/" + filename)

#     browser.quit()

#     return image_preprocess(searchterm)


# def generate_set(class1, class2, class3):
#     return generate_set(class1), generate_set(class2), generate_set(class3)



from textblob import TextBlob


def get_value_instruction(sent):

    # Textblob part of speech identification algorithm
    blob = TextBlob(sent)
    blob.parse()

    # isolating tags of words in instruction
    tags = blob.tags

    decoded = ""

    # if an adjective is present then truth is set to True to activate the
    # correct pipeline
    truth = False
    for x in range(len(tags)):
        if "JJ" in tags[x]:
            truth = True
            break

    # when an adjective exists this pipeline is run
    if truth:
        try:
            for x in range(len(tags)):
                if "JJ" in tags[x]:
                    q = x + 1
                    decoded += sent.split()[x] + "_"

                    # while the word after the adjective is any of these parts
                    # of speech they're added to the instruction final
                    while("VBN" in tags[q] or "VBG" in tags[q] or "NN" in tags[q] or "NNS" in tags[q] or "RB" in tags[q] or ("NNS" in tags[q] and "IN" in tags[q + 1])):
                        decoded += sent.split()[q] + "_"
                        # if an interjection is present then you want to skip
                        # over it
                        if("IN" in tags[q + 1]):
                            decoded += sent.split()[q + 1] + "_"
                            q += 2
                            continue
                        q += 1
                        if q >= len(tags):
                            break

        except BaseException:
            pass

    # if there's no adjective present you want to run this pipeline
    else:
        try:
            # you iterate through the tags and identify certain parts of speech
            for x in range(len(tags)):
                if x < len(tags) - 1:
                    if "IN" in tags[x + 1]:
                        decoded += sent.split()[x] + "_"
                        decoded += sent.split()[x + 1] + "_"
                        x = x + 2
                        continue
                # if any of these parts of speech are in the instruction then
                # you want to extract them. The parts of speech can be found:
                # https://repository.upenn.edu/cgi/viewcontent.cgi?article=1603&context=cis_reports
                if "NN" in tags[x] or "NNS" in tags[x] or "RB" in tags[x] or "VBG" in tags[x] or "VBN" in tags[x]:
                    decoded += sent.split()[x] + "_"
                else:
                    continue
        except BaseException:
            print(x)
            print(tags[x])
            print("Please try re-typing your sentence")

    decoded = decoded[:-1]

    # If it's two words then you just choose the second word: we're assuming two words = predict apples, even if this is false in the example apples red
    # similarity identificatin will still pick up on the right column

    if len(sent.split()) == 2:
        decoded = sent.split()[1]

    return decoded



import os
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'
from libra.modeling.tuner import (tuneReg,
                                  tuneClass,
                                  tuneCNN)
import matplotlib.pyplot as plt
from libra.preprocessing.data_reader import DataReader
from keras.preprocessing.image import ImageDataGenerator
from libra.plotting.generate_plots import (generate_regression_plots,
                                           generate_classification_plots)

import uuid
from PIL import Image
from colorama import Fore, Style

number = 0

counter = 0
currLog = ""

# # allows for all columns to be displayed when printing()
# pd.options.display.width = None


# # clears the log when new process is started up


def clearLog():
    global counter
    global currLog

    currLog = ""
    counter = 0


# logging function that creates hierarchial display of the processes of
# different functions. Copied into different python files to maintain
# global variable parallels


def logger(instruction, found=""):
    '''
    logging function that creates hierarchial display of the processes of
    different functions. Copied into different python files to maintain
    global variables.

    :param instruction: what you want to be displayed
    :param found: if you want to display something found like target column

    '''
    global counter
    if counter == 0:
        print((" " * 2 * counter) + str(instruction) + str(found))
    elif instruction == "->":
        counter = counter - 1
        print(Fore.BLUE + (" " * 2 * counter) +
              str(instruction) + str(found) + (Style.RESET_ALL))
    else:
        print((" " * 2 * counter) + "|- " + str(instruction) + str(found))
        if instruction == "done...":
            print("\n" + "\n")

    counter += 1


def tune_helper(
        model_to_tune=None,
        dataset=None,
        models=None,
        max_layers=10,
        min_layers=2,
        min_dense=32,
        max_dense=512,
        executions_per_trial=3,
        max_trials=1,
        activation='relu',
        loss='categorical_crossentropy',
        metrics='accuracy',
        seed=42,
        objective='val_accuracy',
        generate_plots=True,
        directory='my_dir',
        epochs=10,
        step=32,
        patience=1,
        verbose=0,
        test_size=0.2
):
    '''
    Helper function that calls the appropriate tuning function
    :param instruction: the objective that you want to reduce dimensions to maximize
    :return the updated models dictionary
    '''
    print("")
    logger("Getting target model for tuning...")

    # checks to see which requested model is in the self.models

    # processing for regression feed forward NN
    if model_to_tune == 'regression_ANN':
        logger("Reading in data")
        logger("Tuning model hyperparameters...")
        dataReader = DataReader(dataset)
        data = dataReader.data_generator()
        target = models['regression_ANN']['target']
        target_column = data[models['regression_ANN']['target']]
        data = models['regression_ANN']['preprocesser'].transform(
            data.drop(target, axis=1))
        returned_model, returned_pms, history, X_test, y_test = tuneReg(
            data,
            target_column,
            max_layers=max_layers,
            min_layers=min_layers,
            min_dense=min_dense,
            max_dense=max_dense,
            executions_per_trial=executions_per_trial,
            max_trials=max_trials,
            epochs=epochs,
            activation=activation,
            step=step,
            directory=directory,
            verbose=verbose,
            test_size=test_size
        )
        plots = {}
        logger("->", 'Best Hyperparameters Found: {}'.format(returned_pms.values))
        if generate_plots:
            logger("Generating updated plots")
            init_plots, plot_names = generate_regression_plots(
                history, data, target_column)
            for x in range(len(plot_names)):
                plots[str(plot_names[x])] = init_plots[x]

        models['regression_ANN'] = {
            'id': models['regression_ANN']['id'],
            'model': returned_model,
            'target': target,
            "plots": plots,
            'preprocesser': models['regression_ANN']['preprocesser'],
            'interpreter': models['regression_ANN']['interpreter'],
            'test_data': {'X' : X_test, 'y' : y_test},
            'hyperparameters': returned_pms.values,
            'losses': {
                'training_loss': history.history['loss'],
                'val_loss': history.history['val_loss']}
        }
        logger("Re-stored model under 'regression_ANN' key")

        # processing for classification feed forward NN
    elif model_to_tune == "classification_ANN":
        logger("Reading in data")
        logger("Tuning model hyperparameters...")
        dataReader = DataReader(dataset)
        data = dataReader.data_generator()
        target = models['classification_ANN']['target']
        target_column = data[models['classification_ANN']['target']]
        data = models['classification_ANN']['preprocesser'].transform(
            data.drop(target, axis=1))
        returned_model, returned_pms, history, X_test, y_test = tuneClass(
            data,
            target_column,
            models['classification_ANN']['num_classes'],
            max_layers=max_layers,
            min_layers=min_layers,
            min_dense=min_dense,
            max_dense=max_dense,
            executions_per_trial=executions_per_trial,
            max_trials=max_trials,
            activation=activation,
            loss=loss,
            directory=directory,
            metrics=metrics,
            epochs=epochs,
            step=step,
            verbose=verbose,
            test_size=test_size
        )
        plots = {}
        logger("->", 'Best Hyperparameters Found: {}'.format(returned_pms.values))
        if generate_plots:
            logger("Generating updated plots")
            plots = generate_classification_plots(
                    history, data, target_column, returned_model, X_test, y_test)


        logger("Re-stored model under 'classification_ANN' key")
        models['classification_ANN'] = {
            'id': models['classification_ANN']['id'],
            'model': returned_model,
            'hyperparameters': returned_pms.values,
            'plots': plots,
            'preprocesser': models['classification_ANN']['preprocesser'],
            'interpreter': models['classification_ANN']['interpreter'],
            'test_data': {'X' : X_test, 'y': y_test},
            'target': target,
            'losses': {
                'training_loss': history.history['loss'],
                'val_loss': history.history['val_loss']},
            'accuracy': {
                'training_accuracy': history.history['accuracy'],
                'validation_accuracy': history.history['val_accuracy']
            }
        }

    elif model_to_tune == "convolutional_NN":
        logger("Tuning model hyperparameters...")
        X_train, X_test, height, width, num_classes = get_image_data(models)
        logger('Located image data')
        model, returned_pms, history = tuneCNN(
            X_train,
            X_test,
            height,
            width,
            num_classes,
            executions_per_trial=executions_per_trial,
            max_trials=max_trials,
            seed=seed,
            objective=objective,
            directory=directory,
            patience=patience,
            epochs=epochs,
            verbose=verbose,
            test_size=test_size
        )
        logger("->", "Optimal image size identified: {}".format((height, width, 3)))
        logger('Packaging HyperModel')
        logger("->", 'Best Hyperparameters Found: {}'.format(returned_pms.values))
        logger("Re-stored model under 'convolutional_NN' key")

        models['convolutional_NN'] = {
            'id': models['convolutional_NN']['id'],
            'data_type': models['convolutional_NN']['data_type'],
            'data_path': models['convolutional_NN']['data_path'],
            'data': {'train': X_train, 'test': X_test},
            'shape': models['convolutional_NN']['shape'],
            'model': model,
            'num_classes': models['convolutional_NN']['num_classes'],
            'data_sizes': models['convolutional_NN']['data_sizes'],
            'losses': {
                'training_loss': history.history['loss'],
                'val_loss': history.history['val_loss']},
            'accuracy': {
                'training_accuracy': history.history['accuracy'],
                'validation_accuracy': history.history['val_accuracy']}
        }
    clearLog()
    return models


def save(model, save_model, save_path=os.getcwd()):
    '''
    function used to save a model with its .h5 file and .json instance
    :param modeL; the models dictionary used
    :param save_model, whether you want to save the model
    :param save_path: location where you want to save the model
    '''
    global number
    model_json = model.to_json()
    with open(save_path + "/model" + str(number) + ".json", "w") as json_file:
        json_file.write(model_json)
        # serialize weights to HDF5
        model.save_weights(save_path + "/weights" + str(number) + ".h5")
        logger("->", "Saved model to disk as model" + str(number))
    number = number + 1

def generate_id():
    '''
    function to generate a unique id.
    '''

    return str(uuid.uuid4())


def get_image_data(models):
    '''
    function to get image data from a certain folder specifically for CNN tuning.
    Assumes CNN query has already been called.
    :param data_path: represents the location of the two training/testing folders.
    :param read_mode: represents the type of reading it does: setwise, pathwise, or classwise
    :param training_ratio: represents the size of the training / testing set.
    '''
    training_path = "/proc_training_set"
    testing_path = "/proc_testing_set"

    # read_type = set_distinguisher(data_path, read_mode)['read_mode']

    input_shape = models['convolutional_NN']['shape']
    input_single = (input_shape[0], input_shape[1])
    num_classes = models['convolutional_NN']['num_classes']
    loss_func = ""
    data_path = models['convolutional_NN']['data_path']

    if num_classes > 2:
        loss_func = "categorical_crossentropy"
    elif num_classes == 2:
        loss_func = "binary_crossentropy"

    train_data = ImageDataGenerator(rescale=1. / 255,
                                    shear_range=0.2,
                                    zoom_range=0.2,
                                    horizontal_flip=True)
    test_data = ImageDataGenerator(rescale=1. / 255)

    X_train = train_data.flow_from_directory(data_path + training_path,
                                             target_size=input_single,
                                             color_mode='rgb',
                                             batch_size=(32 if models['convolutional_NN']['data_sizes']['train_size'] >= 32 else 1),
                                             class_mode=loss_func[:loss_func.find("_")]
                                             )
    X_test = test_data.flow_from_directory(data_path + testing_path,
                                           target_size=input_single,
                                           color_mode='rgb',
                                           batch_size=(32 if models['convolutional_NN']['data_sizes']['test_size'] >= 32 else 1),
                                           class_mode=loss_func[:loss_func.find("_")])

    return X_train, X_test, input_shape[0], input_shape[1], num_classes


def get_model_data(self, model):
    '''
    function to print out model dictionary information
    :param modeL: represents the specific model key that you're using to index dictionary
    '''
    if model in self.models:
        return self.models[model].keys()
    else:
        raise Exception(
            "The requested model has not been applied to the client.")

    clearLog()

def get_operators(self, model):
    '''
    gets the operators that were used to preprocess the dataset for prediction.
    :param modeL; the key in the models dictionary.
    '''
    defined = ['plots', 'accuracy', 'losses']
    if model in self.models:
        operations = [
            func +
            "()" for func in self.models[model].keys() if func in defined]
        if len(operations) > 0:
            print(operations)
        else:
            raise Exception(
                "There are no built-in operators defined for this model. Please refer to the models dictionary.")
    else:
        raise Exception(
            "The requested model has not been applied to the client.")


def get_accuracy(self, model):
    '''
    gets the accuracies for a specific model
    :param modeL; the key in the models dictionary.
    '''
    if model in self.models:
        if 'accuracy' in self.models[model].keys():
            return self.models[model]['accuracy']
        elif 'cross_val_score' in self.models[model].keys():
            return {'cross_val_score': self.models[model]['cross_val_score']}
        else:
            raise Exception("Accuracy is not defined for {}".format(model))
    else:
        raise Exception(
            "The requested model has not been applied to the client.")


def get_losses(self, model):
    '''
    gets the losses for that specific model
    :param modeL; the key in the models dictionary.
    '''
    if model in self.models:
        if 'losses' in self.models[model].keys():
            return self.models[model]['losses']
        else:
            raise Exception("Losses are not defined for {}".format(model))
    else:
        raise Exception(
            "The requested model has not been applied to the client.")


def get_target(self, model):
    '''
    gets the target for that specific model if it exists.
    :param modeL; the key in the models dictionary.
    '''
    if model in self.models:
        if 'target' in self.models[model].keys():
            return self.models[model]['target']
        else:
            raise Exception("Target is not defined for {}".format(model))
    else:
        raise Exception(
            "The requested model has not been applied to the client.")


def get_vocab(self, model):
    '''
    gets the vocab that was used in the nlp queries
    :param modeL; the key in the models dictionary.
    '''
    if model in self.models:
        if 'vocabulary' in self.models[model].keys():
            return self.models[model]['vocabulary']
        else:
            raise Exception("Vocabulary is not defined for {}".format(model))
    else:
        raise Exception(
            "The requested model has not been applied to the client.")


def get_plots(self, model="", plot="", save=False):
    '''
    function to get plots and then save them if appropriate
    :param modeL; the key in the models dictionary.
    :param plot: specific plot to get if applicable
    :param save: whether to save the file as a .png
    '''
    # no model or plot specified so plot all
    if model == "" and plot == "":
        model = self.latest_model
        if "plots" in self.models[model].keys():
            for each_plot in self.models[model]['plots']:
                save_and_plot(self, model, each_plot, save)
        else:
            raise Exception("{} does not have plots".format(model))
    # show plots for specified model
    elif model != "" and plot == "":
        if "plots" in self.models[model].keys():
            for each_plot in self.models[model]['plots']:
                save_and_plot(self, model, each_plot, save)
        else:
            raise Exception("{} does not have plots".format(model))
    # show specified plot for specified model
    elif model != "" and plot != "":
        if plot in self.models[model]['plots'].keys():
            theplot = self.models[model]['plots'][plot]
            save_and_plot(self, model, theplot, save)
        else:
            raise Exception("{} is not available for {}".format(plot, model))
    else:
        raise Exception("Invalid plotting input")


# function to save and plot
def save_and_plot(self, modelname, plotname, save):
    '''
    helper function to save the model as a .png
    :param modelname: represents the model from which to save
    :param plotname: specific plot to save
    :param save: whether to save it or not
    '''
    img = self.models[modelname]['plots'][plotname]
    if save:
        img.savefig("{}_{}.png".format(modelname, plotname))
    img.show()




import seaborn as sns
import warnings
from sklearn.metrics import roc_curve, auc
from sklearn.preprocessing import label_binarize
from numpy import interp
import pandas as pd
import sklearn
import matplotlib.pyplot as plt
from sklearn.metrics import confusion_matrix, plot_confusion_matrix, recall_score, precision_score, f1_score, \
    ConfusionMatrixDisplay
import numpy as np
from sklearn.exceptions import DataConversionWarning

warnings.filterwarnings(action='ignore', category=DataConversionWarning)
warnings.filterwarnings("ignore", category=RuntimeWarning)

currLog = ""
counter = 0
number = 0
# current_dir=os.getcw()

# allows for all columns to be displayed when printing()
pd.options.display.width = None


# clears the log when new process is started up


def clearLog():
    global currLog
    global counter

    currLog = ""
    counter = 0


def logger(instruction, found=""):
    '''
    logging function that creates hierarchial display of the processes of
    different functions. Copied into different python files to maintain
    global variables.

    :param instruction: what you want to be displayed
    :param found: if you want to display something found like target column
    '''
    global currLog
    global counter
    if counter == 0:
        currLog += (" " * 2 * counter) + str(instruction) + str(found)
    elif instruction == "->":
        counter = counter - 1
        currLog += (" " * 2 * counter) + str(instruction) + str(found)
    else:
        currLog += (" " * 2 * counter) + "|- " + str(instruction) + str(found)
        if instruction == "done...":
            currLog += "\n" + "\n"

    counter += 1
    print(currLog)
    currLog = ""


# generates all of the plots in clustering


def generate_clustering_plots(
        kmeans,
        dataPandas,
        dataset,
        scatters,
        inertia_sor,
        base_clusters):
    '''
    plotting function that generates all plots for the clustering algorithm

    :param base_clusters: is the number of clusters testing began at
    :param inertia_sor: the array that has all the inertia values
    :param scatters: which scatter plots you want to generate of columns compared to each other
    :param dataset: is the dataset that you want to generate plots for
    :param dataPandas: a pandas version of that same set
    :param kmeans: is the actual kmeans algorithm you're generating for
    :return an array with all of the plots stored as figures inside of it
    '''

    plots = []
    plot_names = []
    # for all of the columns that're present in clustering it plots the
    # columns with each other based on the cluster they're in
    for x in range(len(dataPandas.columns) - 1):
        for y in range(len(dataPandas.columns) - 1):
            if dataPandas.columns[x].replace(
                    " ", "_") + "_vs_" + dataPandas.columns[y].replace(" ", "_") in scatters:
                img = plt.figure()
                plt.scatter(dataset[:, x], dataset[:, y],
                            c=kmeans.labels_, cmap='rainbow')
                plt.xlabel(str(dataPandas.columns[x]))
                plt.ylabel(str(dataPandas.columns[y]))
                plots.append(img)
                plot_names.append(
                    dataPandas.columns[x] +
                    "_vs_" +
                    dataPandas.columns[y])
                plt.close(img)
    return plots, plot_names, elbow_cluster_graph(inertia_sor, base_clusters)


# generates all of the plots for regression
def elbow_cluster_graph(inertia_stor, base_clusters):
    '''
    plotting function that generates the elbow graph

    :param base_clusters: number of clusters you begin testing at
    :param inertia_stor: the array of inertia values
    :return the elbow graph
    '''
    ranged = []
    for i in range(base_clusters, len(inertia_stor) + base_clusters):
        ranged.append(i + 1)

    img = plt.figure()
    plt.plot(ranged, inertia_stor, marker='o')
    # plt.plot(inertia_stor)
    plt.title('Elbow Graph for Clustering')
    plt.ylabel('SSE/Inertia')
    plt.xlabel('Number of Clusters')
    return img


def generate_regression_plots(history, data, label):
    '''
    plotting function that generates regression plots

    :param history: the keras history object
    :return the names and actual regression plots
    '''
    plots = []
    plot_names = []
    # generating plots for loss
    plots.append(plot_loss(history))
    plot_names.append('lossvsval_los')

    return plots, plot_names


def generate_classification_plots(history, data, label, model, X_test, y_test):
    '''
    plotting function that generates classification plots

    :param history: the keras history object
    :return the names and actual classification plots
    '''
    plots = []
    plot_names = []

    # generating plots for accuracy
    plots.append(plot_acc(history))
    plot_names.append('accuracyvsval_accuracy')
    # generating plots for loss
    plots.append(plot_loss(history))
    plot_names.append('lossvsval_los')

    # dynamic way to return all possible plots in case it expands together
    return_plots = {}
    for x in range(len(plots)):
        return_plots[str(plot_names[x])] = plots[x]

    return return_plots


# function to return both val and accuracy plots on one pane


def generate_classification_together(history, data, model, X_test, y_test):
    '''
    unused plotting function that generates multiple plots on one pane when defaulted

    :param history: is the history object of keras
    :param many params: information that is given to the plots for specifications
    :return plots on the same pane
    '''
    plots = []
    plot_names = []

    arrEpochs = []
    # stores all of the history information
    for x in range(len(history.history['loss'])):
        arrEpochs.append(x + 1)

    # the first loss plot on the top
    plt.subplot(2, 1, 1)
    plot_loss(history)
    # the second accuracy plot on the bottom
    plt.subplot(2, 1, 2)
    plot_acc(history)
    plt.show()


def plot_loss(history):
    '''
    plotting function that generates loss plots

    :param history: is the history object from keras
    :return a plot of validation vs training loss
    '''
    img = plt.figure()
    plt.plot(history.history['loss'])
    plt.plot(history.history['val_loss'])
    plt.title('model loss')
    plt.ylabel('loss')
    plt.xlabel('epoch')
    plt.legend(['train', 'test'], loc='upper left')
    return img


def plot_corr(data, col=[]):
    '''
    correlation matrix for plot roc

    :param data: is the data to find correlations for
    :param col: is the specific column you want to find correlations to
    :return correlation matrix
    '''

    img = plt.figure()
    if col:
        data = data.loc[:, data.columns.intersection(['a', 'b'])]
    corr = data.corr()
    sns.heatmap(
        corr,
        vmin=-1, vmax=1, center=0,
        square=True)
    return img


def plot_acc(history):
    '''
    plotting function that generates accuracy plots

    :param history: is the history object from keras
    :return a plot of validation vs training accuracy
    '''
    img = plt.figure()
    plt.plot(history.history['accuracy'])
    plt.plot(history.history['val_accuracy'])
    plt.title('model accuracy')
    plt.ylabel('accuracy')
    plt.xlabel('epoch')
    plt.legend(['train', 'test'], loc='upper left')
    return img


def plot_mc_roc(y_test, y_score, interpreter=None):
    '''
    plotting function that generates roc curves for data given to it.

    :param y_test: is the testing data used
    :param y_score: is the score when the testing data was called
    :param interpreter: is what was used to preprocess
    :return a roc plot
    '''
    lw = 2
    n_classes = len(np.unique(y_test))
    classes = pd.unique(y_test)
    if n_classes != 2:
        y_test = label_binarize(y_test, classes=classes)
        y_score = label_binarize(y_score, classes=classes)
    else:
        n_classes = 1
        y_test = y_test.reshape(-1, 1)
        y_score = y_score.reshape(-1, 1)

    # Compute ROC curve and ROC area for each class
    fpr = dict()
    tpr = dict()
    roc_auc = dict()
    for i in range(n_classes):
        fpr[i], tpr[i], _ = roc_curve(y_test[:, i], y_score[:, i])
        roc_auc[i] = auc(fpr[i], tpr[i])

    # Compute micro-average ROC curve and ROC area
    fpr["micro"], tpr["micro"], _ = roc_curve(y_test.ravel(), y_score.ravel())
    roc_auc["micro"] = auc(fpr["micro"], tpr["micro"])
    # First aggregate all false positive rates
    all_fpr = np.unique(np.concatenate([fpr[i] for i in range(n_classes)]))

    # Then interpolate all ROC curves at this points
    mean_tpr = np.zeros_like(all_fpr)
    for i in range(n_classes):
        mean_tpr += interp(all_fpr, fpr[i], tpr[i])

    # Finally average it and compute AUC
    mean_tpr /= n_classes

    fpr["macro"] = all_fpr
    tpr["macro"] = mean_tpr
    roc_auc["macro"] = sklearn.metrics.auc(fpr["macro"], tpr["macro"])

    # Plot all ROC curves
    img = plt.figure()
    plt.plot(fpr["micro"], tpr["micro"],
             label='micro-average ROC curve (area = {0:0.2f})'
                   ''.format(roc_auc["micro"]),
             color='deeppink', linestyle=':', linewidth=4)

    plt.plot(fpr["macro"], tpr["macro"],
             label='macro-average ROC curve (area = {0:0.2f})'
                   ''.format(roc_auc["macro"]),
             color='navy', linestyle=':', linewidth=4)

    for i in range(n_classes):
        if isinstance(interpreter, dict):
            inverted_interpreter = dict(map(reversed, interpreter.items()))
            plt.plot(fpr[i], tpr[i], lw=lw,
                     label='ROC curve of class {0} (area = {1:0.2f})'
                           ''.format(inverted_interpreter[i], roc_auc[i]))
        else:
            plt.plot(
                fpr[i],
                tpr[i],
                lw=lw,
                label='ROC curve of class {0} (area = {1:0.2f})'
                      ''.format(
                    interpreter.inverse_transform(
                        [[i]])[0],
                    roc_auc[i]))

    plt.plot([0, 1], [0, 1], 'k--', lw=lw)
    plt.xlim([0.0, 1.0])
    plt.ylim([0.0, 1.05])
    plt.xlabel('False Positive Rate')
    plt.ylabel('True Positive Rate')
    plt.title('ROC Curves')
    plt.legend(loc="lower right")
    return img


# Analysis of model
def analyze(client, model=None, save=True, save_model=False):
    '''
    the body of the analyze function in queries.py. Used to generate ROC, confusion matrix etc.
    :param model: is the actual model that you want to analyze for and against
    :param client: is the whole client object :)
    :param save: whether to save into client dictionary
    :param save_model: whether to save as .png
    '''

    if model is None:
        model = client.latest_model

    if not model in client.models:
        raise NameError('Model name is not valid or has not been created yet.')

    plt.clf()
    logger("Analyzing {} for further understanding".format(model))

    modeldict = client.models[model]

    if 'test_data' in modeldict:
        logger("Making predictions for test data")
        data = modeldict['test_data']['X']
        real = modeldict['test_data']['y']
        preds = modeldict['model'].predict(data)
        if model == 'Text Classification':
            logger("Identifying textual relationships and numerical correspondence")
            preds = np.argmax(preds, axis=-1)
        if model == 'classification_ANN':  # formats labels column
            enc = sklearn.preprocessing.LabelEncoder()
            real = modeldict['interpreter'].inverse_transform(
                real).reshape(1, -1)[0]
            preds = modeldict['interpreter'].inverse_transform(
                preds).reshape(1, -1)[0]
            real = enc.fit_transform(real)
            preds = enc.transform(preds)
            logger("Transforming dataset for plotting module")

    if model == 'k_means_clustering':
        logger("Gathering metrics for display: ")
        inertia = modeldict['model'].inertia_
        centers = modeldict['model'].cluster_centers_
        logger(" ", ("Optimal number of clusters: {}".format(str(len(centers)))))
        logger(
            "->",
            ("Sum of squared distance or inertia: {}".format(
                str(inertia))))
        if save:
            modeldict['n_centers'] = len(centers)
            modeldict['centroids'] = centers
            modeldict['inertia'] = inertia
    elif model == 'regression_ANN':
        logger("Gathering metrics for display: ")
        MSE = sklearn.metrics.mean_squared_error(real, preds)
        MAE = sklearn.metrics.mean_absolute_error(real, preds)
        logger(" ", ("MSE on test set: {}".format(str(MSE))))
        logger("->", ("MAE on test set: {}".format(str(MAE))))
        if save:
            modeldict['MSE'] = MSE
            modeldict['MAE'] = MAE
    # classification models
    elif model in ['svm', 'nearest_neighbor', 'decision_tree', 'classification_ANN', 'Text Classification']:
        logger("->", "Plotting ROC curves and creating confusion matrix...")
        if model in ['svm', 'nearest_neighbor',
                     'decision_tree', 'Text Classification']:
            label_source = modeldict['interpreter']
            labels = list(modeldict['interpreter'].keys())
        else:
            label_source = enc
            labels = enc.classes_

        # create roc plots
        roc = plot_mc_roc(real, preds, label_source)

        # create confusion matrices
        cm = confusion_matrix(real, preds)
        cm = ConfusionMatrixDisplay(
            confusion_matrix=cm,
            display_labels=labels).plot()
        cm = cm.figure_

        logger('Investigating potential issues with calculations')
        logger("Gathering metrics for display: ")
        # get accuracy from modeldict
        if model in ['svm', 'nearest_neighbor',
                     'decision_tree']:
            accuracy = modeldict['accuracy']['accuracy_score']
        else:
            accuracy = modeldict['accuracy']['validation_accuracy']
        recall = recall_score(real, preds, average='micro')
        precision = precision_score(real, preds, average='micro')
        f1 = f1_score(real, preds, average='micro')

        logger("->",
               ("Accuracy on test set: {}".format(str((accuracy[-1] if isinstance(accuracy, list) else accuracy)))))
        logger("->", ("Recall on test set: {}".format(str(recall))))
        logger("->", ("Precision on test set: {}".format(str(precision))))
        logger("->", ("F1 Score on test set: {}".format(str(f1))))
        if save:
            if 'plots' not in modeldict:
                modeldict['plots'] = {}
            modeldict['plots']['roc_curve'] = roc
            modeldict['plots']['confusion_matrix'] = cm

            if 'scores' not in modeldict:
                modeldict['scores'] = {}
            modeldict['scores']['recall_score'] = recall
            modeldict['scores']['precision_score'] = precision
            modeldict['scores']['f1_score'] = f1

    else:
        print("further analysis is not supported for {}".format(model))

    if 'plots' in modeldict and model != 'k_means_clustering':
        client.plots(model=model, save=save_model)

    clearLog()



import json
import keras
import numpy as np
from colorama import Fore, Style
import pprint
import pandas as pd
import tensorflow as tf
from tabulate import tabulate
from scipy.spatial.distance import cosine
from sklearn.model_selection import cross_val_score
from pandas import DataFrame
from sklearn.preprocessing import LabelEncoder
from sklearn.metrics import accuracy_score
from sklearn import preprocessing, svm
from sklearn.compose import ColumnTransformer
from sklearn.model_selection import train_test_split
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import OneHotEncoder, StandardScaler
from tensorflow import keras
from tensorflow.python.keras.layers import Dense, Input
from libra.data_generation.dataset_labelmatcher import get_similar_column, get_similar_model
from keras.callbacks import EarlyStopping
from matplotlib import pyplot
from libra.data_generation.grammartree import get_value_instruction
from libra.preprocessing.data_preprocesser import structured_preprocesser
from libra.modeling.prediction_model_creation import get_keras_model_reg
from libra.modeling.prediction_model_creation import get_keras_model_class
from keras.utils import to_categorical
from keras.utils import np_utils
from sklearn.cluster import KMeans
import matplotlib.pyplot as plt

from libra.data_generation import dataGen
from keras.models import Sequential
from keras.layers import (Dense, Conv2D, Flatten)
from os import listdir
from libra.modeling.tuner import tuneReg, tuneClass, tuneCNN
from sklearn.neighbors import KNeighborsClassifier
from sklearn.ensemble import RandomForestRegressor
from sklearn.feature_selection import SelectFromModel
from sklearn import preprocessing, tree
import sys

def initializer(params):
    data = pd.read_csv(params['path_to_set'])
    data.fillna(0, inplace=True)
    params['data'] = data
    return params


def preprocesser(params):
    data = structured_preprocesser(params['data'])
    params['data'] = data


def instruction_identifier(params):
    remove = get_similar_column(
        get_value_instruction(
            params['instruction']),
        params['data'])
    params['y'] = params['data'][remove]
    del params['data'][remove]


def set_splitter(params):
    params['X_train'], params['X_test'], params['y_train'], params['y_test'] = train_test_split(
        params['data'], params['y'], test_size=0.2, random_state=49)


def modeler(params):
    params['maximizer'] = 'loss'
    models = []
    losses = []

    # callback function to store lowest loss value
    params['es'] = EarlyStopping(
        monitor=params['maximizer'],
        mode='min',
        verbose=1,
        patience=5)

    i = 0

    # get the first 3 layer model
    model = get_keras_model_reg(params['data'], i)
    history = model.fit(
        params['X_train'],
        params['y_train'],
        epochs=5,
        validation_data=(
            params['X_test'],
            params['y_test']),
        callbacks=[params['es']])
    models.append(history)

    losses.append(models[i].history[params['maximizer']]
                  [len(models[i].history[params['maximizer']]) - 1])

    # keeps running model and fit functions until the validation loss stops
    # decreasing

    while(all(x > y for x, y in zip(losses, losses[1:]))):
        model = get_keras_model_reg(params['data'], i)
        history = model.fit(
            params['X_train'],
            params['y_train'],
            epochs=5,
            validation_data=(
                params['X_test'],
                params['y_test']),
            callbacks=[params['es']])
        models.append(history)
        losses.append(models[i].history[params['maximizer']]
                      [len(models[i].history[params['maximizer']]) - 1])
        i += 1

    params['models'] = models
    params['losses'] = losses


def plotter(params):
    init_plots, plot_names = generateRegressionPlots(
        params['models'][len(params['models']) - 1], params['data'], params['y'])
    plots = {}
    for x in range(len(plot_names)):
        plots[str(plot_names[x])] = init_plots[x]
    params['plots'] = plots



# Lemmatizer for text
import re
import tensorflow as tf
import spacy
from spacy.lang.en import English
from nltk.corpus import stopwords
from libra.data_generation.dataset_labelmatcher import get_similar_column
from libra.data_generation.grammartree import get_value_instruction


def get_target_values(data, instruction, yLabel):
    # Get target columns
    target = get_similar_column(get_value_instruction(instruction), data)
    X = data[target]
    del data[target]
    # labels
    Y = data[get_similar_column(get_value_instruction(yLabel), data)]
    return X, Y, get_similar_column(get_value_instruction(yLabel), data)


def lemmatize_text(dataset):
    result = []
    nlp = spacy.load('en')
    for text in range(len(dataset)):
        word = ""
        doc = nlp(dataset[text])
        for token in doc:
            if word == "":
                word = token.lemma_
            else:
                word = word + " " + token.lemma_
        result.append(word)
    return result


# Tokenize text

def tokenize_text(dataset):
    nlp = English()
    # Create a Tokenizer with the default settings for English
    # including punctuation rules and exceptions
    tokenizer = nlp.Defaults.create_tokenizer(nlp)
    for i in range(len(dataset)):
        dataset[i] = tokenizer(dataset[i])
    return dataset


# Cleans up text data by removing unnecessary characters (links,
# punctuation, uppercase letters, numbers, whitespace)

def text_clean_up(dataset):
    newDataset = []
    for text in dataset:
        clean_text = re.sub(r'http\S+', '', text)
        punctuation = '!"#$%&()*+-/:;<=>?@[\\]^_`{|}~'
        clean_text = ''.join(
            ch for ch in clean_text if ch not in set(punctuation))
        clean_text = clean_text.lower()
        clean_text = re.sub(r'\d', ' ', clean_text)
        clean_text = ' '.join(clean_text.split())
        clean_text = clean_text.split()
        stops = set(stopwords.words("english"))
        clean_text = [w for w in clean_text if w not in stops]
        clean_text = " ".join(clean_text)
        newDataset.append(fix_slang(clean_text))

    return newDataset


def fix_slang(text):
    text = re.sub(r"i'm", "i am", text)
    text = re.sub(r"he's", "he is", text)
    text = re.sub(r"she's", "she is", text)
    text = re.sub(r"it's", "it is", text)
    text = re.sub(r"that's", "that is", text)
    text = re.sub(r"what's", "that is", text)
    text = re.sub(r"where's", "where is", text)
    text = re.sub(r"how's", "how is", text)
    text = re.sub(r"\'ll", " will", text)
    text = re.sub(r"\'ve", " have", text)
    text = re.sub(r"\'re", " are", text)
    text = re.sub(r"\'d", " would", text)
    text = re.sub(r"\'re", " are", text)
    text = re.sub(r"won't", "will not", text)
    text = re.sub(r"can't", "cannot", text)
    text = re.sub(r"n't", " not", text)
    text = re.sub(r"n'", "ng", text)
    text = re.sub(r"'bout", "about", text)
    text = re.sub(r"'til", "until", text)
    text = re.sub(r"[-()\"#/@;:<>{}`+=~|.!?,]", "", text)
    return text


# text encoder
def encode_text(dataset, text):
    tokenizer = tf.keras.preprocessing.text.Tokenizer(
        num_words=None, filters='!"#$%&()*+,-./:;<=>?@[\\]^_`{|}~\t\n', lower=True,
        split=' ', char_level=False, oov_token=None, document_count=0)
    tokenizer.fit_on_texts(dataset)
    result = tokenizer.texts_to_sequences(text)
    return result

# def decode_sequence(input_seq, encoder_model, target_word_index, decoder_model, reverse_target_word_index,
#                     max_len_summary=50):
#     e_out, e_h, e_c = encoder_model.predict(input_seq)
#
#     target_seq = np.zeros((1, 1))
#
#     target_seq[0, 0] = target_word_index.get('sostok')
#
#     stop_condition = False
#     decoded_sentence = ''
#     while not stop_condition:
#         output_tokens, h, c = decoder_model.predict([target_seq] + [e_out, e_h, e_c])
#
#         # Sample a token
#         sampled_token_index = np.argmax(output_tokens[0, -1, :]) + 1
#         sampled_token = reverse_target_word_index.get(sampled_token_index)
#
#         if sampled_token != 'eostok':
#             print(decoded_sentence)
#             decoded_sentence += ' ' + sampled_token
#
#             # Exit condition: either hit max length or find stop word.
#         if sampled_token == 'eostok' or len(decoded_sentence.split()) >= (max_len_summary - 1):
#            stop_condition = True
#
#         # # Update the target sequence (of length 1).
#         # target_seq = np.zeros((1, 1))
#         # target_seq[0, 0] = sampled_token_index
#
#         # Update internal states
#         e_h, e_c = h, c
#
#     return decoded_sentence
#
#
# def seq2summary(input_seq, target_word_index, reverse_target_word_index):
#     newString = ''
#     for i in input_seq:
#         if (i != 0 and i != target_word_index['sostok']) and i != target_word_index['eostok']:
#             newString = newString + reverse_target_word_index[i] + ' '
#     return newString
#
#
# def seq2text(input_seq, reverse_source_word_index):
#     newString = ''
#     for i in input_seq:
#         if i != 0:
#             newString = newString + reverse_source_word_index[i] + ' '
#     return newString



from regression_split_functions import (
    initializer,
    preprocesser,
    instruction_identifier,
    set_splitter,
    modeler,
    plotter)

#to initialize a testing pipeline, just create a dictionary with both the instruction you want to test and the path to the dataset.
init_params = {
    'instruction': "Predict median house value",
    'path_to_set': './data/housing.csv',
}

reg_pipeline = [initializer,
                mypreprocessor,
                instruction_identifier,
                set_splitter,
                modeler,
                plotter]




[func(init_params) for func in reg_pipeline]


def mypreprocessor():
    return



import numpy as np
import tensorflow as tf
import os


def get_path_column(df):
    for row, col in zip(df.iterrows(), df.columns):
        for entry in row[1]:
            if os.path.exists(entry):
                return col


def load_image(image_path):
    img = tf.io.read_file(image_path)
    img = tf.image.decode_jpeg(img, channels=3)
    img = tf.image.resize(img, (299, 299))
    img = tf.keras.applications.inception_v3.preprocess_input(img)
    return img, image_path


def map_func(img_name, cap):
    img_tensor = np.load(img_name.decode('utf-8') + '.npy')
    return img_tensor, cap


class BahdanauAttention(tf.keras.Model):
    def __init__(self, units):
        super(BahdanauAttention, self).__init__()
        self.W1 = tf.keras.layers.Dense(units)
        self.W2 = tf.keras.layers.Dense(units)
        self.V = tf.keras.layers.Dense(1)

    def call(self, features, hidden):
        # features(CNN_encoder output) shape == (batch_size, 64, embedding_dim)

        # hidden shape == (batch_size, hidden_size)
        # hidden_with_time_axis shape == (batch_size, 1, hidden_size)
        hidden_with_time_axis = tf.expand_dims(hidden, 1)

        # score shape == (batch_size, 64, hidden_size)
        score = tf.nn.tanh(self.W1(features) + self.W2(hidden_with_time_axis))

        # attention_weights shape == (batch_size, 64, 1)
        # you get 1 at the last axis because you are applying score to self.V
        attention_weights = tf.nn.softmax(self.V(score), axis=1)

        # context_vector shape after sum == (batch_size, hidden_size)
        context_vector = attention_weights * features
        context_vector = tf.reduce_sum(context_vector, axis=1)

        return context_vector, attention_weights


class CNN_Encoder(tf.keras.Model):
    # Since you have already extracted the features and dumped it using pickle
    # This encoder passes those features through a Fully connected layer
    def __init__(self, embedding_dim):
        super(CNN_Encoder, self).__init__()
        # shape after fc == (batch_size, 64, embedding_dim)
        self.fc = tf.keras.layers.Dense(embedding_dim)

    def call(self, x):
        x = self.fc(x)
        x = tf.nn.relu(x)
        return x


class RNN_Decoder(tf.keras.Model):
    def __init__(self, embedding_dim, units, vocab_size):
        super(RNN_Decoder, self).__init__()
        self.units = units

        self.embedding = tf.keras.layers.Embedding(vocab_size, embedding_dim)
        self.gru = tf.keras.layers.GRU(self.units,
                                       return_sequences=True,
                                       return_state=True,
                                       recurrent_initializer='glorot_uniform')
        self.fc1 = tf.keras.layers.Dense(self.units)
        self.fc2 = tf.keras.layers.Dense(vocab_size)

        self.attention = BahdanauAttention(self.units)

    def call(self, x, features, hidden):
        # defining attention as a separate model
        context_vector, attention_weights = self.attention(features, hidden)

        # x shape after passing through embedding == (batch_size, 1,
        # embedding_dim)
        x = self.embedding(x)

        # x shape after concatenation == (batch_size, 1, embedding_dim +
        # hidden_size)
        x = tf.concat([tf.expand_dims(context_vector, 1), x], axis=-1)

        # passing the concatenated vector to the GRU
        output, state = self.gru(x)

        # shape == (batch_size, max_length, hidden_size)
        x = self.fc1(output)

        # x shape == (batch_size * max_length, hidden_size)
        x = tf.reshape(x, (-1, x.shape[2]))

        # output shape == (batch_size * max_length, vocab)
        x = self.fc2(x)

        return x, state, attention_weights

    def reset_state(self, batch_size):
        return tf.zeros((batch_size, self.units))


def generate_caption_helper(image,
                            decoder,
                            encoder,
                            tokenizer,
                            image_features_extract_model,
                            max_length=500):
    hidden = decoder.reset_state(batch_size=1)

    temp_input = tf.expand_dims(load_image(image)[0], 0)
    img_tensor_val = image_features_extract_model(temp_input)
    img_tensor_val = tf.reshape(img_tensor_val, (img_tensor_val.shape[0], -1,
                                                 img_tensor_val.shape[3]))

    features = encoder(img_tensor_val)

    dec_input = tf.expand_dims([tokenizer.word_index['<start>']], 0)
    result = []

    for i in range(max_length):
        predictions, hidden, attention_weights = decoder(
            dec_input, features, hidden)

        predicted_id = tf.random.categorical(predictions, 1)[0][0].numpy()
        result.append(tokenizer.index_word[predicted_id])

        if tokenizer.index_word[predicted_id] == '<end>':
            return result

        dec_input = tf.expand_dims([predicted_id], 0)

    return result
